#!/usr/bin/env python3

# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import logging
import datetime

from flask import Flask
from sqlalchemy.sql.expression import asc, desc, cast, or_
from sqlalchemy import Integer, func
import redis

from . import logs
from .models import db, Agent, AgentsGroup, Run, Job, get_setting
from .models import Branch
from .bg import jobs as bg_jobs
from .cloud import cloud
from . import consts
from . import srvcheck
from .. import version
from . import exec_utils
from . import kkrq
from . import utils
from . import dbutils
from . import chops

log = logging.getLogger('watchdog')


def create_app():
    # addresses
    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)

    srvcheck.check_postgresql(db_url)
    srvcheck.wait_for_service('planner', planner_url, 7997)

    logs.setup_logging('watchdog')
    log.info('Kraken Watchdog started, version %s', version.version)

    # Create  Flask app instance
    app = Flask('Kraken Watchdog')

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=watchdog'
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    db.init_app(app)

    # setup sentry
    with app.app_context():
        sentry_url = get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)

    return app


def _exc_handler_with_db_rollback(fnc):
    def inner_function(*args, **kwargs):
        try:
            return fnc(*args, **kwargs)
        except Exception:
            log.exception('IGNORED EXCEPTION')
            db.session.rollback()
            time.sleep(10)
        return None
    return inner_function


@_exc_handler_with_db_rollback
def _check_jobs_if_expired():
    now = utils.utcnow()

    q = Job.query.filter_by(state=consts.JOB_STATE_ASSIGNED)

    job_count = 0
    canceled_count = 0

    for job in q.all():
        log.set_ctx(branch=job.run.flow.branch_id, flow_kind=job.run.flow.kind, flow=job.run.flow_id, run=job.run.id, job=job.id)

        job_count += 1
        if not job.assigned:
            log.warning('job %s assigned but no assign time', job)
            log.reset_ctx()
            continue

        timeout = job.timeout if job.timeout else consts.DEFAULT_JOB_TIMEOUT
        duration = now - job.assigned
        if duration > datetime.timedelta(seconds=timeout):
            log.warning('time %ss for job %s expired, canceling', timeout, job)
            note = 'job expired after %ss' % timeout
            exec_utils.cancel_job(job, note, consts.JOB_CMPLT_SERVER_TIMEOUT)
            canceled_count += 1

        log.reset_ctx()

    if job_count > 0:
        log.info('canceled jobs:%d / all:%d', canceled_count, job_count)


@_exc_handler_with_db_rollback
def _check_jobs_if_missing_agents():
    groups_jobs = {}  # keep here groups that are needed and one, any job

    q = Job.query.filter_by(covered=False, deleted=None, state=consts.JOB_STATE_QUEUED)
    for job in q.all():
        ag = job.agents_group
        if (ag.deployment and
            ag.deployment['method'] in [consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2,
                                        consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE,
                                        consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM,
                                        consts.AGENT_DEPLOYMENT_METHOD_K8S,
                                        ]):
            groups_jobs[ag.id] = job.id

    for ag_id in groups_jobs:
        kkrq.enq_neck(bg_jobs.spawn_new_agents, ag_id)
    if groups_jobs:
        n = len(groups_jobs)
        grps = list(groups_jobs.keys())[:5]
        job = groups_jobs.popitem()[1]
        log.info('enqueued spawning new agents for %d groups eg:%s with no agents but with waiting jobs, eg: %s',
                 n, grps, job)


def _check_jobs():
    _check_jobs_if_expired()
    _check_jobs_if_missing_agents()


@_exc_handler_with_db_rollback
def _check_runs_timeout():
    now = utils.utcnow()

    q = Run.query.filter_by(finished=None)
    q = q.filter(Run.state != consts.RUN_STATE_MANUAL)  # TODO: manual runs will require timeouting as well when jobs are started

    for run in q.all():
        timeout = run.stage.schema.get('timeout', consts.DEFAULT_RUN_TIMEOUT)
        if run.started:
            begin = run.started
        else:
            begin = run.created
        end_time = begin + datetime.timedelta(seconds=timeout)
        if end_time > now:
            # no timeout yet
            continue

        log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

        note = 'run %s timed out, deadline was: %s' % (run, str(end_time))
        log.info(note)
        run.note = note

        # cancel any pending jobs
        canceled_jobs_count = 0
        for job in run.jobs:
            if job.state != consts.JOB_STATE_COMPLETED:
                exec_utils.cancel_job(job, note, consts.JOB_CMPLT_SERVER_TIMEOUT)
                canceled_jobs_count += 1

        # if there is no pending jobs then complete run now
        if canceled_jobs_count == 0:
            exec_utils.complete_run(run, now)

        log.reset_ctx()


@_exc_handler_with_db_rollback
def _check_runs_missing_history_analysis():
    now = utils.utcnow()
    three_months = now - datetime.timedelta(days=90)
    twenty_mins = now - datetime.timedelta(minutes=20)

    q = Run.query.filter(Run.finished > three_months)
    q = q.filter(Run.finished < twenty_mins)
    q = q.filter(Run.state != consts.RUN_STATE_PROCESSED)
    q = q.order_by(asc(Run.finished))

    visited_branches = set()
    for run in q.all():
        if run.flow.branch_id in visited_branches:
            continue

        log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

        log.info('run with missing history %s', run)
        visited_branches.add(run.flow.branch_id)
        kkrq.enq_neck(bg_jobs.analyze_results_history, run.id)

        log.reset_ctx()


def _check_runs():
    _check_runs_timeout()
    _check_runs_missing_history_analysis()


def _cancel_job_and_unassign_agent(agent):
    exec_utils.cancel_job(agent.job, 'agent not alive', consts.JOB_CMPLT_AGENT_NOT_ALIVE)
    job = agent.job
    agent.job = None
    job.agent = None
    db.session.commit()


@_exc_handler_with_db_rollback
def _check_agents_keep_alive():
    now = utils.utcnow()
    five_mins_ago = now - datetime.timedelta(seconds=consts.AGENT_TIMEOUT)
    ten_mins_ago = now - datetime.timedelta(seconds=consts.SLOW_AGENT_TIMEOUT)

    q = Agent.query
    q = q.filter_by(disabled=False, deleted=None)
    q = q.filter(Agent.last_seen < five_mins_ago)

    for a in q.all():
        log.reset_ctx()
        log.set_ctx(agent=a.id)
        # in case of cloud machines check for 10mins, not 5mins
        ag = dbutils.find_cloud_assignment_group(a)
        if (ag and
            ag.deployment['method'] in [consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2,
                                        consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE,
                                        consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM,
                                        consts.AGENT_DEPLOYMENT_METHOD_K8S]):
            if a.last_seen > ten_mins_ago:
                continue

        a.disabled = True
        a.status_line = 'agent was not seen for last 5 minutes, disabled'
        log.info('agent %s not seen for 5 minutes, disabled', a)
        db.session.commit()

        if a.job:
            _cancel_job_and_unassign_agent(a)


def _destroy_and_delete_if_outdated(agent, depl, method):
    destroy = False

    if method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
        # if for some reason k8s pod still exists after completing the job
        # then destroy it after 3 mins
        destruction_after_time = 3
    else:
        if 'destruction_after_time' in depl:
            destruction_after_time = int(depl['destruction_after_time'])
        else:
            destruction_after_time = 10  # default if somehow not set

    # check if idle time after job passed, then destroy VM
    if destruction_after_time > 0:
        max_idle_time = datetime.timedelta(seconds=60 * destruction_after_time)

        q = Job.query.filter_by(agent_used=agent)
        q = q.filter(Job.finished.isnot(None))
        q = q.filter(Job.finished > agent.created)
        q = q.order_by(desc(Job.finished))
        last_job = q.first()
        now = utils.utcnow()
        if last_job:
            dt = now - last_job.finished
            if dt >= max_idle_time:
                log.info('agent: %s, timed out %s > %s - destroying it, last job %s',
                         agent, dt, max_idle_time, last_job)
                destroy = True
            else:
                log.info('agent: %s, not yet timed out %s < %s - skipped', agent, dt, max_idle_time)
        elif now - agent.created >= max_idle_time:
            log.info('agent: %s, timed out %s - destroying it, created at %s vs now %s', agent, max_idle_time, agent.created, now)
            destroy = True
        else:
            log.info('agent: %s, no last job and not idle enough - skipped', agent)
    else:
        log.info('agent: %s, destruction_after_time is 0 - skipped', agent)

    # check if number of executed jobs on VM is reached, then destroy VM
    if not destroy:
        if 'destruction_after_jobs' in depl and int(depl['destruction_after_jobs']) > 0:
            max_jobs = int(depl['destruction_after_jobs'])
            q = Job.query.filter_by(agent_used=agent, state=consts.JOB_STATE_COMPLETED)
            q = q.filter(Job.finished > agent.created)
            jobs_num = q.count()
            if jobs_num >= max_jobs:
                log.info('agent: %s, max jobs reached %d/%d', agent, jobs_num, max_jobs)
                # TODO: remove it later, for debugging only
                #for j in q.all():
                #    log.info('  job: %s', j)
                destroy = True
            else:
                log.info('agent: %s, max jobs not reached yet %d/%d - skipped', agent, jobs_num, max_jobs)
        else:
            log.info('agent: %s, destruction_after_jobs is 0 - skipped', agent)

    # if machine mark for destruction then schedule it
    if destroy:
        log.info('disabling and destroying machine with agent %s', agent)
        agent.disabled = True
        db.session.commit()
        kkrq.enq(bg_jobs.destroy_machine, agent.id)
        return True

    return False


def _delete_if_missing_in_cloud(ag, agent):
    if not agent.extra_attrs or 'instance_id' not in agent.extra_attrs:
        log.warning('agent:%d, no instance id in extra_attrs', agent.id)
        return False

    exists = True
    try:
        exists = cloud.check_if_machine_exists(ag, agent)
    except Exception:
        pass

    if not exists:
        dbutils.delete_agent(agent)
        log.info('deleted dangling agent %s', agent)
        return True

    return False


@_exc_handler_with_db_rollback
def _check_agents_to_destroy():
    # look for agents that are not deleted, still available for use ie. authorized
    # and without currently running job
    # that have some deployment method configured
    q = Agent.query.filter_by(deleted=None, authorized=True, job=None)
    q = q.join('agents_groups', 'agents_group')
    q = q.filter(AgentsGroup.deployment.isnot(None))
    # TODO ECS
    q = q.filter(or_(cast(AgentsGroup.deployment['method'], Integer) == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2,
                     cast(AgentsGroup.deployment['method'], Integer) == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM,
                     cast(AgentsGroup.deployment['method'], Integer) == consts.AGENT_DEPLOYMENT_METHOD_K8S))

    outdated_count = 0
    dangling_count = 0
    all_count = 0
    for agent in q.all():
        log.reset_ctx()
        log.set_ctx(agent=agent.id)

        all_count += 1
        ag = dbutils.find_cloud_assignment_group(agent)

        if not ag:
            log.error('missing ag in agent %s', agent)
            continue

        if not ag.deployment:
            log.error('missing deployment in ag %s', ag)
            continue

        method = consts.AGENT_DEPLOYMENT_METHOD_MANUAL
        try:
            method, depl = ag.get_deployment()
        except Exception:
            pass
        if method not in [consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2,
                          consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM,
                          consts.AGENT_DEPLOYMENT_METHOD_K8S]:
            log.error('unsupported deployment method %s', ag.deployment['method'])
            continue
        deleted = _destroy_and_delete_if_outdated(agent, depl, method)
        if deleted:
            outdated_count += 1
            if agent.job:
                _cancel_job_and_unassign_agent(agent)
            continue

        deleted = _delete_if_missing_in_cloud(ag, agent)
        if deleted:
            dangling_count += 1
            if agent.job:
                _cancel_job_and_unassign_agent(agent)
            continue

    log.reset_ctx()

    if outdated_count > 0:
        log.info('all agents:%d, destroyed and deleted %d VM instances and agents',
                 all_count, outdated_count)
    if dangling_count > 0:
        log.info('deleted %d dangling agents without any VM instance', dangling_count)

    return all_count, outdated_count, dangling_count


@_exc_handler_with_db_rollback
def _check_machines_with_no_agent():
    # look for AWS EC2 or Azure VM machines that do not have agents in database
    # and destroy such machines
    q = AgentsGroup.query
    q = q.filter_by(deleted=None)
    q = q.filter(AgentsGroup.deployment.isnot(None))

    all_groups = 0
    aws_ec2_groups = 0
    azure_vm_groups = 0
    k8s_groups = 0

    for ag in q.all():
        all_groups += 1

        if ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
            aws_ec2_groups += 1
        elif ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
            azure_vm_groups += 1
        elif ag.deployment['method'] == consts.AGENT_DEPLOYMENT_METHOD_K8S:
            k8s_groups += 1
        else:
            continue

        counts = cloud.cleanup_dangling_machines(ag)

        instances = counts[0]
        terminated_instances = counts[1]
        assigned_instances = counts[2]
        orphaned_instances = counts[3]
        orphaned_terminated_instances = counts[4]

        if (instances + terminated_instances + assigned_instances +
            orphaned_instances + orphaned_terminated_instances > 0):
            log.info('group: %s, instances:%d, already-terminated:%d, still-assigned:%d, orphaned:%d, terminated-orphaned:%d',
                     ag,
                     instances,
                     terminated_instances,
                     assigned_instances,
                     orphaned_instances,
                     orphaned_terminated_instances)
    if aws_ec2_groups + azure_vm_groups + k8s_groups > 0:
        log.info('machines with no agent by groups: aws:%d, azure vm:%d, k8s:%d / all:%d',
                 aws_ec2_groups, azure_vm_groups, k8s_groups, all_groups)


@_exc_handler_with_db_rollback
def _check_agents_counts():
    q = db.session.query(Agent.authorized, func.count(Agent.authorized))
    q = q.filter_by(deleted=None)
    q = q.group_by(Agent.authorized)
    counts = q.all()

    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_KRAKEN_DB)

    for authorized, cnt in counts:
        if authorized:
            rds.set('authorized-agents', cnt)
        else:
            rds.set('non-authorized-agents', cnt)


def _check_agents():
    _check_agents_keep_alive()
    _check_agents_to_destroy()
    _check_machines_with_no_agent()
    _check_agents_counts()


@_exc_handler_with_db_rollback
def _check_for_errors_in_logs():
    ch = chops.get_clickhouse()

    now = utils.utcnow()
    start_date = now - datetime.timedelta(hours=1)

    query = "select count(*) from logs where level = 'ERROR' and time > %(start_date)s;"
    rows = ch.execute(query, {'start_date': start_date})
    errors_count = rows[0][0]

    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_KRAKEN_DB)

    rds.set('error-logs-count', errors_count)
    #log.info('updated errors count to %s', errors_count)


@_exc_handler_with_db_rollback
def _run_branch_retention_policy():
    log.info('branches logs retention')
    ch = chops.get_clickhouse()

    q = Branch.query
    q = q.filter_by(deleted=None)

    for branch in q.all():
        log.reset_ctx()
        log.set_ctx(branch=branch.id)

        if branch.retention_policy:
            rp = branch.retention_policy
            months = [rp['ci_logs'], rp['dev_logs']]
        else:
            months = [6, 3]

        for flow_kind in [0, 1]:
            query = "ALTER TABLE logs DELETE WHERE branch = %(branch_id)s AND flow_kind = %(flow_kind)s AND time < (now() - toIntervalMonth(%(months)s))"
            params = dict(branch_id=branch.id,
                          flow_kind=flow_kind,
                          months=months[flow_kind])
            resp = ch.execute(query, params)
            log.info('DELETED LOGS %s', resp)
            # TODO: trace number of deleted logs

    log.reset_ctx()


def _main_loop():
    t0 = time.time()
    t0_log_errs = t0_jobs = t0_runs = t0_agents = t0
    t0_branch_retention_policy = t0


    while True:
        # check jobs every 5 seconds
        dt = time.time() - t0_jobs
        if dt > 5:
            _check_jobs()
            t0_jobs = time.time()

        # check for error in logs every 15 seconds
        dt = time.time() - t0_log_errs
        if dt > 15:
            _check_for_errors_in_logs()
            t0_log_errs = time.time()

        # check runs every 30 seconds
        dt = time.time() - t0_runs
        if dt > 30:
            _check_runs()
            t0_runs = time.time()

        # check agents every 30 seconds
        dt = time.time() - t0_agents
        if dt > 30:
            _check_agents()
            t0_agents = time.time()

        # run retention policy every 24 h
        dt = time.time() - t0_branch_retention_policy
        if dt > 60 * 60 * 24:
            _run_branch_retention_policy()
            t0_branch_retention_policy = time.time()

        time.sleep(1)

def main():
    app = create_app()

    with app.app_context():
        while True:
            try:
                _main_loop()
            except Exception:
                log.exception('IGNORED EXCEPTION')
                db.session.rollback()
                time.sleep(10)


if __name__ == "__main__":
    main()



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .bg import jobs as bg_jobs
from . import consts
from . import kkrq

log = logging.getLogger(__name__)


def trigger_run(stage_id, flow_kind=consts.FLOW_KIND_CI, reason=None):
    logging.basicConfig(format=consts.LOG_FMT, level=logging.INFO)

    args = (stage_id, flow_kind, reason)

    log.info('trigger run for stage %s', stage_id)
    kkrq.enq_neck(bg_jobs.trigger_run, *args, ignore_args=[2])


def refresh_schema_repo(stage_id):
    logging.basicConfig(format=consts.LOG_FMT, level=logging.INFO)

    log.info('refresh stage %s schema from repo', stage_id)
    kkrq.enq_neck(bg_jobs.refresh_schema_repo, stage_id, None, ignore_args=[1])



# Copyright 2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from contextlib import contextmanager

from redis import Redis

## vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
## copied from https://github.com/pycasbin/sqlalchemy-adapter/blob/606a631b1704d76c5ca0f83064158604851dc17f/casbin_sqlalchemy_adapter/adapter.py
## this is under apache 2.0 license


import casbin
from casbin import persist
from sqlalchemy import or_
from werkzeug.exceptions import Forbidden

from .models import db, CasbinRule
from . import utils


log = logging.getLogger(__name__)


class Filter:
    ptype = []
    v0 = []
    v1 = []
    v2 = []
    v3 = []
    v4 = []
    v5 = []


class Adapter(persist.Adapter, persist.adapters.UpdateAdapter):
    """the interface for Casbin adapters."""

    def __init__(self, filtered=False):
        self._filtered = filtered

    @contextmanager
    def _session_scope(self):
        """Provide a transactional scope around a series of operations."""
        try:
            yield
            db.session.commit()
        except Exception as e:
            db.session.rollback()
            raise e

    def load_policy(self, model):
        """loads all policy rules from the storage."""
        with self._session_scope():
            lines = CasbinRule.query.all()
            for line in lines:
                persist.load_policy_line(str(line), model)

    def is_filtered(self):
        return self._filtered

    def load_filtered_policy(self, model, flter) -> None:
        """loads all policy rules from the storage."""
        with self._session_scope():
            q = CasbinRule.query
            q = self.filter_query(q, flter)
            filtered = q.all()

            for line in filtered:
                persist.load_policy_line(str(line), model)
            self._filtered = True

    def filter_query(self, querydb, flter):
        for attr in ("ptype", "v0", "v1", "v2", "v3", "v4", "v5"):
            if len(getattr(flter, attr)) > 0:
                querydb = querydb.filter(
                    getattr(CasbinRule, attr).in_(getattr(flter, attr))
                )
        return querydb.order_by(CasbinRule.id)

    def _save_policy_line(self, ptype, rule):
        with self._session_scope():
            line = CasbinRule(ptype=ptype)
            for i, v in enumerate(rule):
                setattr(line, "v{}".format(i), v)
            # session.add(line)

    def save_policy(self, model):
        """saves all policy rules to the storage."""
        with self._session_scope():
            CasbinRule.query.delete()
            for sec in ["p", "g"]:
                if sec not in model.model.keys():
                    continue
                for ptype, ast in model.model[sec].items():
                    for rule in ast.policy:
                        self._save_policy_line(ptype, rule)
        return True

    def add_policy(self, sec, ptype, rule):  # pylint: disable=unused-argument
        """adds a policy rule to the storage."""
        self._save_policy_line(ptype, rule)

    def add_policies(self, sec, ptype, rules):  # pylint: disable=unused-argument
        """adds a policy rules to the storage."""
        for rule in rules:
            self._save_policy_line(ptype, rule)

    def remove_policy(self, sec, ptype, rule):  # pylint: disable=unused-argument
        """removes a policy rule from the storage."""
        with self._session_scope():
            q = CasbinRule.query.filter_by(ptype=ptype)
            for i, v in enumerate(rule):
                q = q.filter(getattr(CasbinRule, "v{}".format(i)) == v)
            r = q.delete()

        return r > 0

    def remove_policies(self, sec, ptype, rules):  # pylint: disable=unused-argument
        """remove policy rules from the storage."""
        if not rules:
            return
        with self._session_scope():
            q = CasbinRule.query.filter_by(ptype=ptype)
            rules = zip(*rules)
            for i, rule in enumerate(rules):
                q = q.filter(
                    or_(getattr(CasbinRule, "v{}".format(i)) == v for v in rule)
                )
            q.delete()

    def remove_filtered_policy(self, sec, ptype, field_index, *field_values):  # pylint: disable=unused-argument
        """removes policy rules that match the filter from the storage.
        This is part of the Auto-Save feature.
        """
        with self._session_scope():
            q = CasbinRule.query.filter_by(ptype=ptype)

            if not (0 <= field_index <= 5):
                return False
            if not (1 <= field_index + len(field_values) <= 6):
                return False
            for i, v in enumerate(field_values):
                if v != "":
                    v_value = getattr(CasbinRule, "v{}".format(field_index + i))
                    q = q.filter(v_value == v)
            r = q.delete()

        return r > 0

    def update_policy(
        self, sec: str, ptype: str, old_rule: [str], new_rule: [str]
    ) -> None:  # pylint: disable=unused-argument
        """
        Update the old_rule with the new_rule in the database (storage).

        :param sec: section type
        :param ptype: policy type
        :param old_rule: the old rule that needs to be modified
        :param new_rule: the new rule to replace the old rule

        :return: None
        """

        with self._session_scope():
            q = CasbinRule.query.filter_by(ptype=ptype)

            # locate the old rule
            for index, value in enumerate(old_rule):
                v_value = getattr(CasbinRule, "v{}".format(index))
                q = q.filter(v_value == value)

            # need the length of the longest_rule to perform overwrite
            longest_rule = old_rule if len(old_rule) > len(new_rule) else new_rule
            old_rule_line = q.one()

            # overwrite the old rule with the new rule
            for index in range(len(longest_rule)):
                if index < len(new_rule):
                    exec(f"old_rule_line.v{index} = new_rule[{index}]")
                else:
                    exec(f"old_rule_line.v{index} = None")

    def update_policies(
        self,
        sec: str,
        ptype: str,
        old_rules: [
            [str],
        ],
        new_rules: [
            [str],
        ],
    ) -> None:
        """
        Update the old_rules with the new_rules in the database (storage).

        :param sec: section type
        :param ptype: policy type
        :param old_rules: the old rules that need to be modified
        :param new_rules: the new rules to replace the old rules

        :return: None
        """
        for i in range(len(old_rules)):
            self.update_policy(sec, ptype, old_rules[i], new_rules[i])

    def update_filtered_policies(
        self, sec, ptype, new_rules: [[str]], field_index, *field_values
    ) -> [[str]]:
        """update_filtered_policies updates all the policies on the basis of the filter."""

        flter = Filter()
        flter.ptype = ptype

        # Creating Filter from the field_index & field_values provided
        for i in range(len(field_values)):
            if field_index <= i and i < field_index + len(field_values):
                setattr(flter, f"v{i}", field_values[i - field_index])
            else:
                break

        self._update_filtered_policies(new_rules, flter)

    def _update_filtered_policies(self, new_rules, flter) -> [[str]]:
        """_update_filtered_policies updates all the policies on the basis of the filter."""

        with self._session_scope():

            # Load old policies
            q = CasbinRule.query.filter_by(ptype=flter.ptype)
            q = self.filter_query(q, flter)
            old_rules = q.all()

            # Delete old policies
            self.remove_policies("p", flter.ptype, old_rules)

            # Insert new policies
            self.add_policies("p", flter.ptype, new_rules)

            # return deleted rules
            return old_rules


## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

class CasbinWatcher:

    REDIS_CHANNEL_NAME = 'casbin-watcher'

    def __init__(self, redis_addr):
        self.stale = False
        self.redis_addr = redis_addr
        if not self.redis_addr:
            return
        self.watcher_thread = threading.Thread(target=self._watcher, daemon=True)
        self.watcher_thread.start()

    def _watcher(self):
        host, port = utils.split_host_port(self.redis_addr, 6379)
        r = Redis(host, port)
        p = r.pubsub()
        p.subscribe(self.REDIS_CHANNEL_NAME)
        log.info("Waiting for casbin policy updates...")
        while r:
            # wait 20 seconds to see if there is a casbin update
            try:
                message = p.get_message(timeout=20)
            except Exception:
                log.exception("Casbin watcher failed to get message from redis due to")
                p.close()
                r = None
                break

            if message and message.get('type') == "message":
                log.info("Casbin policy update identified, message: %s", message)
                self.stale = True

    def notify(self):
        if not self.redis_addr:
            return
        host, port = utils.split_host_port(self.redis_addr, 6379)
        r = Redis(host, port)
        r.publish(self.REDIS_CHANNEL_NAME, 'updated')


ROLE_SUPERADMIN = "superadmin"
ROLE_VIEWER = 'viewer'
ROLE_PWRUSR = 'pwrusr'
ROLE_ADMIN = 'admin'


enforcer = None

def init(redis_addr=None):
    global enforcer
    if enforcer:
        return

    adapter = Adapter()

    # Create a config model policy
    model_txt = """
        [request_definition]
        r = sub, obj, act

        [policy_definition]
        p = sub, obj, act

        [role_definition]
        g = _, _
        g2 = _, _

        [policy_effect]
        e = some(where (p.eft == allow))

        [matchers]
        m = (g(r.sub, p.sub) && r.obj == p.obj && r.act == p.act) || g2(r.sub, 'superadmin') || r.sub == '1'
    """
    model = casbin.Enforcer.new_model(text=model_txt)

    # Create enforcer from adapter and config policy
    enforcer = casbin.Enforcer(model, adapter)

    enforcer.my_watcher = CasbinWatcher(redis_addr)


def check(token_info, obj, act, msg):
    user = token_info['sub']
    sub2 = str(user.id)
    obj2 = str(obj)
    act2 = str(act)
    log.info('check access sub(user):%s obj:%s act:%s',
             sub2, obj2, act2)

    if user.details and not user.details.get('enabled', True):
        raise Forbidden('User %s is disabled' % user.name)

    if enforcer.my_watcher.stale:
        enforcer.load_policy()
        enforcer.my_watcher.stale = False

    if not enforcer.enforce(sub2, obj2, act2):
        raise Forbidden(msg)


def get_user_roles(user):
    # force reloading policies from db
    enforcer.load_policy()

    resp = {}

    if user.name == 'admin':
        resp['superadmin'] = True
    else:
        policies = enforcer.get_filtered_named_grouping_policy("g2", 0, str(user.id))
        superadmin = False
        for _, role in policies:
            if role == ROLE_SUPERADMIN:
                superadmin = True
                break

        resp['superadmin'] = superadmin

    policies = enforcer.get_filtered_named_grouping_policy("g", 0, str(user.id))
    projects = {}
    for _, role in policies:
        parts = role.split('-')
        proj_role = parts[0]
        proj_id = int(parts[1][1:])
        projects[proj_id] = proj_role
    resp['projects'] = projects

    return resp


def notify_policy_change():
    enforcer.my_watcher.notify()



"""added unique index for address of non-deleted agents

Revision ID: 9d8d8713c1ad
Revises: 19396769e240
Create Date: 2021-05-22 07:04:18.171520

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '9d8d8713c1ad'
down_revision = '19396769e240'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    conn.execute("CREATE UNIQUE INDEX ix_agents_address_not_deleted ON agents (address) WHERE deleted is NULL;")


def downgrade():
    op.drop_index('ix_agents_address_not_deleted')



#!/usr/bin/env python3

# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import json
import logging

import redis

from . import logs
from . import consts
from . import srvcheck
from . import utils
from .. import version
from .bg import jobs as bg_jobs
from . import kkrq

log = logging.getLogger('qneck')


def _check_jobs(waiting_jobs, executing_jobs):
    # check of jobs then were sent for execution
    # if they are maybe completed
    completed = []
    for key, job in executing_jobs.items():
        status = job.get_status()
        if status in ['finished', 'failed']:
            # TODO: what about failed retries?
            completed.append(key)

    # go through completed jobs and 1) remove them
    # from execution jobs and 2) if there is the same
    # job in waiting jobs then run it
    for key in completed:
        del executing_jobs[key]

        vals = waiting_jobs.get(key, None)
        if vals is None:
            continue

        func, args = vals
        job = kkrq.enq(func, *args)
        executing_jobs[key] = job

        del waiting_jobs[key]


def _main_loop():
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    srvcheck.wait_for_service('redis', redis_addr, 6379)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_RQ_DB)

    ps = rds.pubsub()
    ps.subscribe('qneck')

    waiting_jobs = {}
    executing_jobs = {}

    last_check = time.time()

    while True:
        try:
            # get the job from redis pubsub channel
            # wait for it up to 1 second, if there is
            # nothing then None is returned
            msg = ps.get_message(timeout=1)

            # if received not interesting message
            # then check what is happening with jobs
            if not msg or msg['type'] != 'message':
                if time.time() - last_check > 5:
                    _check_jobs(waiting_jobs, executing_jobs)
                    last_check = time.time()
                continue

            log.info('qneck: received msg: %s', msg)

            # unpack message
            data = json.loads(msg['data'])
            func_name = data['func']
            args = tuple(data['args'])
            ignore_args = data['ignore_args']
            func = getattr(bg_jobs, func_name)

            if ignore_args:
                key = [func]
                for idx, a in enumerate(args):
                    if idx in ignore_args:
                        continue
                    key.append(a)
                key = tuple(key)
            else:
                key = (func, *args)

            # if the same jobs as received one is waiting
            # then drop this one - we keep only one waiting job
            if key in waiting_jobs:
                log.info('qneck: skipped %s%s as it is already in queue', func, args)
                continue

            # if the same job as received one is being executed
            # then put it in waiting jobs - it will be executed
            # current one is finished
            if key in executing_jobs:
                waiting_jobs[key] = (func, args)
                log.info('qneck: queued %s%s as it is already being executed', func, args)
                continue

            # if the same job as received job is not waiting
            # nor it is being executed then run the received job
            job = kkrq.enq(func, *args)
            executing_jobs[key] = job

        except Exception:
            log.exception('IGNORED EXCEPTION')
            time.sleep(1)


def main():
    logs.setup_logging('rq')
    log.info('Kraken QNeck started, version %s', version.version)

    try:
        _main_loop()
    except Exception:
        log.exception('IGNORED EXCEPTION')
        time.sleep(10)


if __name__ == "__main__":
    main()



# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import logging
import xmlrpc.client

import pytimeparse
import dateutil.parser
import RestrictedPython
from RestrictedPython import compile_restricted
from RestrictedPython import limited_builtins

from . import consts
from . import schemaval
from .models import Secret


log = logging.getLogger(__name__)


class SchemaError(Exception):
    pass


class SchemaCodeContext:
    def __init__(self, branch_name, context):
        self.branch_name = branch_name
        for k, v in context.items():
            setattr(self, k, v)


def execute_schema_code(branch, schema_code, context=None):
    # TODO: use starlark-go for executing schema code
    # for now RestrictedPython is used
    try:
        byte_code = compile_restricted(schema_code, '<inline>', 'exec')
    except Exception as e:
        log.exception('schema compilation exception')
        msg = 'compilation error: ' + str(e)
        raise SchemaError(msg) from e

    my_locals = {}
    my_globals = {'__builtins__': limited_builtins,
                  '_getattr_': RestrictedPython.Guards.safer_getattr,
                  '_getiter_': RestrictedPython.Eval.default_guarded_getiter,
                  '_iter_unpack_sequence_': RestrictedPython.Guards.guarded_iter_unpack_sequence}


    try:
        exec(byte_code, my_globals, my_locals)  # pylint: disable=exec-used
    except Exception as e:
        log.exception('schema compilation exception')
        msg = 'compilation error: ' + str(e)
        raise SchemaError(msg) from e

    my_globals.update(my_locals)
    if context is None:
        context = {
            'is_ci': True,
            'is_dev': False,
        }
    ctx = SchemaCodeContext(branch.name, context)
    my_globals['ctx'] = ctx

    my_locals2 = {}
    try:
        exec('schema = stage(ctx)', my_globals, my_locals2)  # pylint: disable=exec-used
    except Exception as e:
        log.exception('schema compilation exception')
        msg = 'compilation error: ' + str(e)
        raise SchemaError(msg) from e
    schema = my_locals2['schema']

    # validate generated schema
    error = schemaval.validate(schema)
    if error:
        raise SchemaError(error)

    return schema


def check_and_correct_stage_schema(branch, stage_name, schema_code, context=None):
    if not schema_code:
        schema_code = '''def stage(ctx):
    return {
        "parent": "root",
        "triggers": {
            "parent": True,
        },
        "parameters": [],
        "configs": [],
        "jobs": [{
            "name": "hello world",
            "steps": [{
                "tool": "shell",
                "cmd": "echo 'hello world'"
            }],
            "environments": [{
                "system": "any",
                "agents_group": "all",
                "config": "default"
            }]
        }]
    }'''
    log.info('schema_code %s', schema_code)

    # execute schema code
    try:
        schema = execute_schema_code(branch, schema_code, context)
    except SchemaError:
        raise
    except Exception:
        log.exception('unkown error in schema execution')
        raise

    # fill missing parts in schema
    if 'jobs' not in schema:
        schema['jobs'] = []

    if 'configs' not in schema:
        schema['configs'] = []

    if 'parent' not in schema or schema['parent'] == '':
        schema['parent'] = 'root'

    if 'triggers' not in schema or schema['triggers'] == {}:
        schema['triggers'] = {'parent': True}

    if 'parameters' not in schema:
        schema['parameters'] = []

    # check parent in schema
    if schema['parent'] != 'root':
        found = False
        for s in branch.stages:
            if s.deleted:
                continue
            if schema['parent'] == s.name and stage_name != s.name:
                found = True
                break
        if not found:
            raise SchemaError('Cannot find parent stage %s' % schema['parent'])

    # check job_names and secrets
    job_names = set()
    for job in schema['jobs']:
        # check names
        if job['name'] in job_names:
            raise SchemaError("Two jobs with the same name '%s'" % job['name'])

        job_names.add(job['name'])

        # check secrets
        for step in job['steps']:
            for field, value in step.items():
                if field in ['access-token', 'ssh-key']:
                    secret = Secret.query.filter_by(project=branch.project, name=value).one_or_none()
                    if secret is None:
                        raise SchemaError("Secret '%s' does not exist" % value)
                    if field == 'access-token' and secret.kind != consts.SECRET_KIND_SIMPLE:
                        raise SchemaError("Type of '%s' secret should be Simple" % value)
                    if field == 'ssh-key' and secret.kind != consts.SECRET_KIND_SSH_KEY:
                        raise SchemaError("Type of '%s' secret should be SSH Username & Key" % value)

    # TODO: check if git url is valid according to giturlparse
    return schema_code, schema


def prepare_secrets(run):
    secrets = {}
    for s in run.stage.branch.project.secrets:
        if s.deleted:
            continue
        if s.kind == consts.SECRET_KIND_SSH_KEY:
            name = "KK_SECRET_USER_" + s.name
            secrets[name] = s.data['username']
            name = "KK_SECRET_KEY_" + s.name
            secrets[name] = s.data['key']
        elif s.kind == consts.SECRET_KIND_SIMPLE:
            name = "KK_SECRET_SIMPLE_" + s.name
            secrets[name] = s.data['secret']

    return secrets


def substitute_val(val, args):
    if isinstance(val, dict):
        return substitute_vars(val, args)
    if isinstance(val, list):
        list2 = []
        list2_masked = []
        for e in val:
            e2, e2_masked = substitute_val(e, args)
            list2.append(e2)
            list2_masked.append(e2_masked)
        return list2, list2_masked
    if not isinstance(val, str):
        return val, val

    val_masked = val
    secret_present = False
    for var in re.findall(r'#{[A-Za-z0-9_ ]+}', val):
        name = var[2:-1]
        if name in args:
            arg_val = args[name]
            if  not isinstance(arg_val, str):
                raise Exception("value '%s' of '%s' should have string type but has '%s'" % (str(arg_val), name, str(type(arg_val))))
            val = val.replace(var, arg_val)
            if name.startswith('KK_SECRET_'):
                val_masked = val_masked.replace(var, '******')
                secret_present = True
            else:
                val_masked = val_masked.replace(var, arg_val)

    if secret_present:
        return val, val_masked
    return val, val


def substitute_vars(fields, args):
    new_fields = {}
    new_fields_masked = {}
    for f, val in fields.items():
        val, val_masked = substitute_val(val, args)
        new_fields[f] = val
        new_fields_masked[f] = val_masked
    return new_fields, new_fields_masked


def prepare_new_planner_triggers(stage_id, new_triggers, prev_triggers, triggers):
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
    planner = xmlrpc.client.ServerProxy(planner_url, allow_none=True)

    log.info('stage: %d, triggers: new: %s, old: %s', stage_id, new_triggers, prev_triggers)

    # set up interval trigger
    if 'interval' in new_triggers:
        interval = int(pytimeparse.parse(new_triggers['interval']))
        if prev_triggers is None or 'interval' not in prev_triggers or 'interval_planner_job' not in triggers:
            args = (stage_id, consts.FLOW_KIND_CI, dict(reason='interval'))
            job = planner.add_job('kraken.server.pljobs:trigger_run', 'interval', args, None,
                                  None, None, None, None, None, None, False, dict(seconds=interval))
            triggers['interval_planner_job'] = job['id']
            log.info('added new interval trigger - interval:%ds: next: %s', interval, job['next_run_time'])
        else:
            prev_interval = int(pytimeparse.parse(prev_triggers['interval']))
            if interval != prev_interval:
                job = planner.reschedule_job(triggers['interval_planner_job'], 'interval', dict(seconds=interval))
                log.info('rescheduled interval trigger - interval:%ds: next: %s', interval, job['next_run_time'])
    elif 'interval_planner_job' in triggers:
        planner.remove_job(triggers['interval_planner_job'])
        del triggers['interval_planner_job']
        log.info('deleted interval trigger')

    # set up date trigger
    if 'date' in new_triggers:
        run_date = dateutil.parser.parse(new_triggers['date'])
        if prev_triggers is None or 'date' not in prev_triggers or 'date_planner_job' not in triggers:
            args = (stage_id, consts.FLOW_KIND_CI, dict(reason='date'))
            job = planner.add_job('kraken.server.pljobs:trigger_run', 'date', args, None,
                                  None, None, None, None, None, None, False, dict(run_date=str(run_date)))
            triggers['date_planner_job'] = job['id']
            log.info('added new date trigger - date:%s: next: %s', str(run_date), job['next_run_time'])
        else:
            prev_run_date = dateutil.parser.parse(prev_triggers['date'])
            if run_date != prev_run_date:
                job = planner.reschedule_job(triggers['date_planner_job'], 'date', dict(run_date=str(run_date)))
                log.info('rescheduled date trigger - date:%s: next: %s', str(run_date), job['next_run_time'])
    elif 'date_planner_job' in triggers:
        planner.remove_job(triggers['date_planner_job'])
        del triggers['date_planner_job']

    # set up cron trigger
    if 'cron' in new_triggers:
        cron_rule = new_triggers['cron']
        if prev_triggers is None or 'cron' not in prev_triggers or 'cron_planner_job' not in triggers:
            minutes, hours, days, months, dow = cron_rule.split()
            args = (stage_id, consts.FLOW_KIND_CI, dict(reason='cron'))
            job = planner.add_job('kraken.server.pljobs:trigger_run', 'cron', args, None,
                                  None, None, None, None, None, None, False,
                                  dict(minute=minutes, hour=hours, day=days, month=months, day_of_week=dow))
            triggers['cron_planner_job'] = job['id']
            log.info('added new cron trigger - rule:%ds: next: %s', cron_rule, job['next_run_time'])
        else:
            prev_cron_rule = prev_triggers['cron']
            if cron_rule != prev_cron_rule:
                minutes, hours, days, months, dow = cron_rule.split()
                job = planner.reschedule_job(triggers['cron_planner_job'], 'cron',
                                             dict(minute=minutes, hour=hours, day=days, month=months, day_of_week=dow))
                log.info('rescheduled cron trigger - rule:%s: next: %s', cron_rule, job['next_run_time'])
    elif 'cron_planner_job' in triggers:
        planner.remove_job(triggers['cron_planner_job'])
        del triggers['cron_planner_job']

    # set up repo interval trigger
    if 'repo' in new_triggers:
        interval = int(pytimeparse.parse(new_triggers['repo']['interval']))
        if prev_triggers is None or 'repo' not in prev_triggers or 'repo_interval_planner_job' not in triggers:
            args = (stage_id, consts.FLOW_KIND_CI, dict(reason='repo_interval'))
            job = planner.add_job('kraken.server.pljobs:trigger_run', 'repo_interval', args, None,
                                  None, None, None, None, None, None, False, dict(seconds=interval))
            triggers['repo_interval_planner_job'] = job['id']
            log.info('added new repo interval trigger - interval:%ds: next: %s', interval, job['next_run_time'])
        else:
            prev_interval = int(pytimeparse.parse(prev_triggers['repo']['interval']))
            if interval != prev_interval:
                job = planner.reschedule_job(triggers['repo_interval_planner_job'], 'interval', dict(seconds=interval))
                log.info('rescheduled repo interval trigger - interval:%ds: next: %s', interval, job['next_run_time'])
    elif 'repo_interval_planner_job' in triggers:
        planner.remove_job(triggers['repo_interval_planner_job'])
        del triggers['repo_interval_planner_job']
        log.info('deleted repo interval trigger')

    if triggers == {}:
        triggers['parent'] = True



"""added retention_policy to branch

Revision ID: cea0e48abb5a
Revises: 5c16623f4d55
Create Date: 2023-02-16 08:06:50.864851

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'cea0e48abb5a'
down_revision = '5c16623f4d55'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('branches', sa.Column('retention_policy', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('branches', 'retention_policy')



"""added systems table and host_info, user_attrs column to agent table

Revision ID: cb7654d57bca
Revises: 466de256799e
Create Date: 2020-10-11 20:07:55.477525

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.expression import bindparam

# revision identifiers, used by Alembic.
revision = 'cb7654d57bca'
down_revision = '466de256799e'
branch_labels = None
depends_on = None


def upgrade():
    sys_tbl = op.create_table('systems',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=150), nullable=True),
                    sa.Column('executor', sa.Unicode(length=20), nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('name', 'executor', name='uq_system_name_executor'),
    )
    op.add_column('agents', sa.Column('host_info', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('agents', sa.Column('user_attrs', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.add_column('jobs', sa.Column('system_id', sa.Integer(), nullable=True))
    op.create_foreign_key('fk_systems_jobs', 'jobs', 'systems', ['system_id'], ['id'])

    conn = op.get_bind()

    # prepare system records for insertion
    jobs = conn.execute("SELECT id,system FROM jobs")
    systems = []
    jobs_map = {}
    for j_id, j_system in jobs.fetchall():
        print(j_id, j_system)
        if not j_system:
            j_system = 'local^fake'

        if j_system in jobs_map:
            jobs_map[j_system].append({'_id': j_id})
        else:
            parts = j_system.split('^')
            if len(parts) == 2:
                executor, name = parts
            elif len(parts) == 1:
                executor = 'local'
                name = parts[0]
                j_system = '%s^%s' % (executor, name)
                if j_system in jobs_map:
                    jobs_map[j_system].append({'_id': j_id})
                    continue
            else:
                raise Exception('not enough parts: %s' % str(parts))
            systems.append({'name': name, 'executor': executor})
            jobs_map[j_system] = [{'_id': j_id}]

    # insert all system records
    op.bulk_insert(sys_tbl, systems)

    # get all systems id and prepare list of job updates
    systems = conn.execute("SELECT id,name,executor FROM systems")
    sys_map = {}
    jobs_updts = []
    for s_id, s_name, s_executor in systems.fetchall():
        key = '%s^%s' % (s_executor, s_name)
        sys_map[key] = s_id

        for upd in jobs_map[key]:
            upd['system_id'] = s_id
            jobs_updts.append(upd)

    if jobs_updts:
        # update all jobs
        jobs_tbl = sa.sql.table(
            'jobs',
            sa.Column('id', sa.Integer(), nullable=False),
            sa.Column('system_id', sa.Integer(), nullable=True),
            # Other columns not needed for the data migration
        )
        stmt = jobs_tbl.update()
        stmt = stmt.where(jobs_tbl.c.id == bindparam('_id'))
        stmt = stmt.values({
            'system_id': bindparam('system_id')
        })
        conn.execute(stmt, jobs_updts)

    op.drop_column('jobs', 'system')


def downgrade():
    #op.add_column('jobs', sa.Column('system', sa.VARCHAR(length=200), autoincrement=False, nullable=True))
    op.drop_constraint('fk_systems_jobs', 'jobs', type_='foreignkey')
    op.drop_column('jobs', 'system_id')
    op.drop_column('agents', 'user_attrs')
    op.drop_column('agents', 'host_info')
    op.drop_table('systems')



"""added label to run

Revision ID: 466de256799e
Revises: d2b23f410d02
Create Date: 2020-09-26 20:36:07.063973

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '466de256799e'
down_revision = 'd2b23f410d02'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('runs', sa.Column('label', sa.Unicode(length=200), nullable=True))


def downgrade():
    op.drop_column('runs', 'label')



"""added authorized to Executor

Revision ID: 67d4f5a5ee98
Revises: 0ad7507aa6d0
Create Date: 2020-01-06 18:50:52.689999

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '67d4f5a5ee98'
down_revision = '0ad7507aa6d0'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('executors', sa.Column('authorized', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('executors', 'authorized')



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import datetime

import pytz

log = logging.getLogger(__name__)


def utcnow():
    return datetime.datetime.now(pytz.utc)


def split_host_port(addr, default_port):
    parts = addr.split(':')
    host = parts[0]
    if len(parts) == 1:
        if default_port is None:
            raise Exception("format of address '%s' is incorrect" % addr)
        port = default_port
    else:
        port = int(parts[1])
    return host, port



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import json
import copy

import jsonschema
from sqlalchemy.sql.expression import asc

from .models import Tool


# pylint: disable=line-too-long

data1 = {
    "parent": "Tarball",
    "triggers": {
        "parent": True
    },
    "parameters": [],
    "configs": [],
    "jobs": [
        {
            "name": "build",
            "timeout": 5000,
            "steps": [
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "echo \"${GOOGLE_KEY}\" | docker login -u _json_key --password-stdin https://eu.gcr.io",
                    "env": {
                        "GOOGLE_KEY": "#{KK_SECRET_SIMPLE_google_key}"
                    }
                },
                {
                    "tool": "shell",
                    "cmd": "rake build_all",
                    "cwd": "kraken",
                    "timeout": 600,
                    "env": {
                        "kk_ver": "0.#{KK_FLOW_SEQ}"
                    }
                },
                {
                    "tool": "shell",
                    "cmd": "rake publish_docker",
                    "cwd": "kraken",
                    "timeout": 1500,
                    "env": {
                        "kk_ver": "0.#{KK_FLOW_SEQ}"
                    }
                },
                {
                    "tool": "artifacts",
                    "source": [
                        "kraken-docker-compose-*.yaml",
                        ".env"
                    ],
                    "cwd": "kraken",
                    "public": True
                }
            ],
            "environments": [
                {
                    "system": "krakenci/bld-kraken",
                    "executor": "docker",
                    "agents_group": "all",
                    "config": "c1"
                }
            ]
        }
    ],
    "notification": {
        "slack": {
            "channel": "kk-results"
        },
        "email": "godfryd@gmail.com",
        "github": {
            "credentials": "#{KK_SECRET_SIMPLE_gh_status_creds}"
        }
    }
}

data2 = {
    "parent": "Build",
    "triggers": {
        "manual": True
    },
    "parameters": [],
    "configs": [],
    "jobs": [
        {
            "name": "publish on github",
            "steps": [
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "public": True,
                    "source": "kraken-docker-compose-0.#{KK_FLOW_SEQ}.yaml",
                    "cwd": "kraken"
                },
                {
                    "tool": "shell",
                    "cmd": "rake kk_ver=0.#{KK_FLOW_SEQ} github_release",
                    "cwd": "kraken",
                    "env": {
                        "GITHUB_TOKEN": "#{KK_SECRET_SIMPLE_github_token}"
                    }
                }
            ],
            "environments": [
                {
                    "system": "krakenci/bld-kraken",
                    "executor": "docker",
                    "agents_group": "all",
                    "config": "default"
                }
            ]
        }
    ]
}

data3 = {
    "parent": "Tarball",
    "triggers": {
        "parent": True
    },
    "parameters": [],
    "configs": [],
    "jobs": [
        {
            "name": "pylint agent",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo apt-get install -y --no-install-recommends python3-setuptools",
                    "timeout": 300
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install pylint"
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install -r requirements.txt",
                    "cwd": "kraken/agent"
                },
                {
                    "tool": "shell",
                    "cmd": "cp server/kraken/server/logs.py server/kraken/server/consts.py agent/kraken/agent",
                    "cwd": "kraken"
                },
                {
                    "tool": "pylint",
                    "rcfile": "pylint.rc",
                    "modules_or_packages": "agent/kraken/agent",
                    "cwd": "kraken"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "c1"
                }
            ]
        },
        {
            "name": "pylint server",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo apt-get install -y --no-install-recommends python3-setuptools",
                    "timeout": 300
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install poetry"
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "echo 'version = \"0.0\"' > version.py",
                    "cwd": "kraken/server/kraken"
                },
                {
                    "tool": "shell",
                    "cmd": "poetry install -n --no-root",
                    "cwd": "kraken/server",
                    "timeout": 300
                },
                {
                    "tool": "shell",
                    "cmd": "pwd && ls -al",
                    "cwd": "kraken/server"
                },
                {
                    "tool": "pylint",
                    "pylint_exe": "poetry run pylint",
                    "rcfile": "../pylint.rc",
                    "modules_or_packages": "kraken/server",
                    "cwd": "kraken/server"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "c1"
                }
            ]
        },
        {
            "name": "ng lint",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo DEBIAN_FRONTEND=noninteractive apt-get install -yq --no-install-recommends git nodejs npm || ps axf",
                    "timeout": 300
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "npm install",
                    "cwd": "kraken/ui",
                    "timeout": 240
                },
                {
                    "tool": "nglint",
                    "cwd": "kraken/ui"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "c1"
                }
            ]
        },
        {
            "name": "cloc",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo apt-get install -y --no-install-recommends cloc || ps axf",
                    "timeout": 300
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "cloc",
                    "not-match-f": "(package-lock.json|pylint.rc)",
                    "exclude-dir": "docker-elk",
                    "cwd": "kraken"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "c1"
                }
            ]
        }
    ],
    "notification": {
        "slack": {
            "channel": "kk-results"
        },
        "email": "godfryd@gmail.com",
        "github": {
            "credentials": "#{KK_SECRET_SIMPLE_gh_status_creds}"
        }
    }
}

data4 = {
    "parent": "root",
    "triggers": {
        "parent": True
    },
    "parameters": [],
    "configs": [],
    "flow_label": "0.#{KK_FLOW_SEQ}",
    "run_label": "a.#{KK_CI_DEV_RUN_SEQ}",
    "jobs": [
        {
            "name": "tarball",
            "steps": [
                {
                    "tool": "git",
                    "checkout": "https://github.com/Kraken-CI/kraken.git",
                    "branch": "master"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zcvf kraken.tar.gz kraken"
                },
                {
                    "tool": "artifacts",
                    "source": "kraken.tar.gz"
                }
            ],
            "environments": [
                {
                    "system": "any",
                    "agents_group": "all",
                    "config": "default"
                }
            ]
        }
    ],
    "notification": {
        "slack": {
            "channel": "kk-results"
        },
        "email": "godfryd@gmail.com"
    }
}

data5 = {
    "parent": "Tarball",
    "triggers": {
        "parent": True
    },
    "parameters": [],
    "configs": [],
    "jobs": [
        {
            "name": "pytest agent",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo apt-get install -y python3-pip || ps axf",
                    "timeout": 300
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install pytest"
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install -r requirements.txt",
                    "cwd": "kraken/agent"
                },
                {
                    "tool": "shell",
                    "cmd": "cp consts.py logs.py ../../../agent/kraken/agent",
                    "cwd": "kraken/server/kraken/server"
                },
                {
                    "tool": "pytest",
                    "params": "-vv",
                    "cwd": "kraken/agent"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "default"
                }
            ]
        },
        {
            "name": "pytest server",
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sudo apt update && sudo apt-get install -y python3-pip || ps axf",
                    "timeout": 300
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install pytest"
                },
                {
                    "tool": "artifacts",
                    "action": "download",
                    "source": "kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "tar -zxf kraken.tar.gz"
                },
                {
                    "tool": "shell",
                    "cmd": "sudo pip3 install poetry"
                },
                {
                    "tool": "shell",
                    "cmd": "poetry install",
                    "cwd": "kraken/server"
                },
                {
                    "tool": "pytest",
                    "params": "-vv",
                    "cwd": "kraken/server"
                }
            ],
            "environments": [
                {
                    "system": "krakenci/ubuntu:20.04",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "default"
                }
            ]
        }
    ],
    "notification": {
        "slack": {
            "channel": "kk-results"
        },
        "email": "godfryd@gmail.com",
        "github": {
            "credentials": "#{KK_SECRET_SIMPLE_gh_status_creds}"
        }
    }
}

data6 = {
    "parent": "Unit Tests",
    "triggers": {
        "parent": True,
        "cron": "1 * * * *",
        "interval": "10m"
    },
    "parameters": [],
    "configs": [
        {
            "name": "c1",
            "p1": "1",
            "p2": "3"
        },
        {
            "name": "c2",
            "n3": "33",
            "t2": "asdf"
        }
    ],
    "jobs": [
        {
            "name": "make dist",
            "steps": [
                {
                    "tool": "git",
                    "checkout": "https://github.com/frankhjung/python-helloworld.git",
                    "branch": "master"
                },
                {
                    "tool": "pytest",
                    "params": "tests/testhelloworld.py",
                    "cwd": "python-helloworld"
                }
            ],
            "environments": [
                {
                    "system": "all",
                    "agents_group": "all",
                    "config": "c1"
                }
            ]
        }
    ]
}

data7 = {
    "parent": "root",
    "triggers": {
        "repo": {
            "url": "https://github.com/Kraken-CI/kraken.git",
            "branch": "master",
            "interval": "20m"
        }
    },
    "parameters": [],
    "configs": [],
    "jobs": [
        {
            "name": "timeout",
            "timeout": 30,
            "steps": [
                {
                    "tool": "shell",
                    "cmd": "sleep 1000"
                }
            ],
            "environments": [
                {
                    "system": "centos-7",
                    "agents_group": "all",
                    "config": "default"
                }
            ]
        }
    ]
}

data8 = {
    "parent": "root",
    "triggers": {
        "repo": {
            "repos": [{
                "url": "https://github.com/Kraken-CI/kraken",
                "branch": "master"
            }, {
                "url": "https://github.com/Kraken-CI/workflow-examples",
                "branch": "master"
            }],
            "interval": "20m"
        }
    },
    "parameters": [
        {
            "name": "COUNT",
            "type": "string",
            "default": "10",
            "description": "Number of tests to generate"
        }
    ],
    "jobs": [
        {
            "name": "random tests",
            "timeout": 100000,
            "steps": [
                {
                    "tool": "rndtest",
                    "count": "#{COUNT}"
                }
            ],
            "environments": [
                {
                    "system": "ubuntu:19.10",
                    "agents_group": "docker",
                    "executor": "docker",
                    "config": "default"
                }
            ]
        }
    ],
    "notification": {
        "slack": {
            "channel": "kk-results"
        },
        "email": "godfryd@gmail.com"
    }
}

SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Schema of Kraken Workflow Schema",
    "description": "This is a schema that defines format of Kraken Workflow Schema.",

    "type": "object",
    "additionalProperties": False,
    "properties": {
        "parent": {
            "description": "A name of the parent stage. `'root'` if there is no parent. This allows chaining stages.",
            "type": "string"
        },
        "triggers": {
            "description": "One or more triggers that cause starting a new run of this stage.",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "parent": {
                    "description": "A parent stage to current one or 'root'. It allows defining a chain of stages.",
                    "type": "boolean"
                },
                "interval": {
                    "type": "string"
                },
                "date": {
                    "type": "string"
                },
                "cron": {
                    "description": "A parent stage to current one or 'root'. It allows defining a chain of stages.",
                    "type": "string"
                },
                "repo": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "url": {
                            "type": "string"
                        },
                        "branch": {
                            "type": "string"
                        },
                        "repos": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "additionalProperties": False,
                                "properties": {
                                    "url": {
                                        "type": "string"
                                    },
                                    "branch": {
                                        "type": "string"
                                    }
                                }
                            }
                        },
                        "interval": {
                            "oneOf": [{
                                "type": "integer",
                                "minimum": 1
                            }, {
                                "type": "string"
                            }]
                        },
                        "git_cfg": {
                            "description": "Git config keys and values passed to -c of the clone command.",
                            "additionalProperties": True,
                            "properties": {
                            }
                        }
                    }
                },
                "manual": {
                    "type": "boolean"
                }
            }
        },
        "parameters": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "type": {
                        "type": "string"
                    },
                    "default": {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    }
                }
            }
        },
        "configs": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": {
                    "type": "string"
                }
            }
        },
        "flow_label": {
            "description": "A custom label format for flows.",
            "type": "string"
        },
        "run_label": {
            "description": "A custom label format for runs.",
            "type": "string"
        },
        "jobs": {
            "description": "A list of jobs that are executed in the run. Jobs are executed in parallel.",
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "name": {
                        "description": "A name of the job.",
                        "type": "string"
                    },
                    "timeout": {
                        "description": "An initial timeout of the job. If there are more than 10 historical succeded jobs then timeout is estimated automatically.",
                        "type": "integer",
                        "minimum": 30
                    },
                    "steps": {
                        "description": "An array of steps that are executed by an agent. Each step has indicated tool that is executing it. Steps are executed in given order.",
                        "type": "array",
                        "items": {
                            "type": "object",
                        }
                    },
                    "environments": {
                        "description": "It defines the surroundings of a job execution.",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": False,
                            "required": ["system", "agents_group"],
                            "properties": {
                                "agents_group": {
                                    "description": "A name of agents group. An agent from this group will be used to execute the job. There is a special built-in group, `'all'` that gathers all agents.",
                                    "type": "string"
                                },
                                "executor": {
                                    "description": "An executor that agent is using to execute a job.",
                                    "enum": ["local", "docker", "lxd"],
                                    "default": "local",
                                    "type": "string"
                                },
                                "system": {
                                    "description": "An operating system name or list of such names that should be used for job execution. If this is a list then the number of job instances is multiplied by numer of systems - each instance has its system. There is a special system name, `'any'`, that ignores system selection by jobs scheduler.",
                                    "oneOf": [{
                                        "type": "string"
                                    }, {
                                        "type": "array",
                                        "items": {
                                            "type": "string"
                                        }
                                    }]
                                },
                                "config": {
                                    "description": "Not implemented yet.",
                                    "type": "string"
                                }
                            }
                        }
                    }
                }
            }
        },
        "notification": {
            "description": "Notification allows for configuring a notification means that are used to pass an information about stage's run result. There are several communication methods supported.",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "email":  {
                    "description": "It sends run results to indicated email address.",
                    "type": "string"
                },
                "slack": {
                    "description": "It sends run results to indicated Slack channel.",
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "channel": {
                            "description": "Selected Slack channel",
                            "type": "string"
                        }
                    }
                },
                "github": {
                    "description": "It sends run results to associated pull request page on GitHub.",
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "credentials": {
                            "description": "Credentials (user:password) that allows access to the project on GitHub.",
                            "type": "string"
                        }
                    }
                }
            }
        }
    },
    "required": ["parent", "triggers"]
}


def get_schema():
    q = Tool.query
    q = q.filter_by(deleted=None)
    q = q.order_by(asc(Tool.created))

    tools = []
    prev_tool_name = None
    for tool in q.all():
        if tool.name != prev_tool_name:
            prev_tool_name = tool.name
            tools.append((tool, tool.name))

            name = '%s@%s' % (tool.name, tool.version)
            tools.append((tool, name))
        else:
            name = '%s@%s' % (tool.name, tool.version)
            tools.append((tool, name))

    tools_schemas = {}
    for tool, name_ver in tools:
        fields = {
            "tool": {
                "description": tool.description,
                "const": name_ver
            },
            "attempts": {
                "description": "A number of times the step is retried if if it returns error.",
                "default": 1,
                "type": "integer"
            },
            "sleep_time_after_attempt": {
                "description": "A sleep time between subsequent execution attempts.",
                "default": 0,
                "type": "integer"
            },
            "name": {
                "description": "A name of the step, displayed in web UI.",
                "default": "",
                "type": "string"
            }
        }

        if "properties" not in tool.fields:
            raise Exception("Schema of '%s' tool does not have 'properties' field." % name_ver)

        for fname, fdef in tool.fields["properties"].items():
            fields[fname] = fdef

        required_fields = tool.fields.get("required", [])
        if not isinstance(required_fields, list):
            raise Exception("'required' field in '%s' tool fields definitions should have list value, not %s." % (
                name_ver, type(required_fields)))

        td = {
            "if": { "properties": { "tool": { "const": name_ver } } },
            "then": {
                "additionalProperties": tool.fields.get("additionalProperties", False),
                "required": ["tool"] + required_fields,
                "properties": fields
            },
            "else": {
                "additionalProperties": False,
                "properties": {
                }
            }
        }

        tools_schemas[name_ver] = td

    return copy.deepcopy(SCHEMA), tools_schemas


def _ordinal(n):
    return "%d%s" % (n, "tsnrhtdd"[( n // 10 % 10 != 1) * (n % 10 < 4) * n % 10::4])


def validate(instance):
    schema, tools_schemas = get_schema()

    # check main parts
    try:
        jsonschema.validate(instance=instance, schema=schema)
    except jsonschema.exceptions.ValidationError as ex:
        return str(ex)

    # check steps
    for j_idx, j in enumerate(instance['jobs']):
        for s_idx, s in enumerate(j['steps']):
            where = 'in the %s job %s, in the %s step' % (_ordinal(j_idx + 1), j['name'], _ordinal(s_idx + 1))

            if 'tool' not in s:
                return "Missing 'tool' field %s" % where

            if s['tool'] not in tools_schemas:
                return "Unknown tool '%s' %s" % (s['tool'], where)

            tool_schema = tools_schemas[s['tool']]
            try:
                jsonschema.validate(instance=s, schema=tool_schema)
            except jsonschema.exceptions.ValidationError as ex:
                return str(ex)

    return None


def test():
    for data in [data1, data2, data3, data4, data5, data6, data7, data8]:
        try:
            jsonschema.validate(instance=data, schema=SCHEMA)
        except jsonschema.exceptions.ValidationError as ex:
            print('problems: %s' % str(ex))


def dump_to_json():
    with open('kraken.schema.json', 'w') as fp:
        json.dump(SCHEMA, fp)


if __name__ == '__main__':
    if sys.argv[1] == 'test':
        test()
    elif sys.argv[1] == 'json':
        dump_to_json()



"""move webhook to Project

Revision ID: 818e190ca142
Revises: d45087b62027
Create Date: 2019-12-27 07:38:17.002835

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '818e190ca142'
down_revision = 'd45087b62027'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('projects', sa.Column('webhooks', postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.drop_column('stages', 'webhooks')


def downgrade():
    op.add_column('stages', sa.Column('webhooks', postgresql.JSONB(astext_type=sa.Text()), autoincrement=False, nullable=True))
    op.drop_column('projects', 'webhooks')



# Copyright 2020-2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import logging
import tempfile
import subprocess

import furl

from . import minioops

log = logging.getLogger(__name__)


def _run(cmd, check=True, cwd=None, capture_output=False, text=None, secret=None):
    if secret:
        cmd2 = cmd.replace(secret, '******')
    else:
        cmd2 = cmd
    log.info("execute '%s' in '%s'", cmd2, cwd)
    p = subprocess.run(cmd, shell=True, check=check, cwd=cwd, capture_output=capture_output, text=text)
    return p


def _retrieve_git_repo(tmpdir, repo_url, git_cfg, mc, minio_bucket, minio_repo_bundle_path):
    # retrieve git repo bundle from minio
    repo_bundle_path = os.path.join(tmpdir, 'repo.bundle')
    log.info('try to retrieve repo bundle %s / %s -> %s', minio_bucket, minio_repo_bundle_path, repo_bundle_path)
    try:
        mc.fget_object(minio_bucket, minio_repo_bundle_path, repo_bundle_path)
        bundle_present = True
    except Exception as e:
        if 'Object does not exist' in str(e):
            log.info('repo bundled not archived yet')
        else:
            log.exception('retriving repo from bundle failed, skipping it')
        bundle_present = False

    # prepare git config
    cfg_params = []
    for k, v in git_cfg.items():
        cfg_params.append('-c %s=%s' % (k, v))
    cfg_params = ' '.join(cfg_params)

    repo_dir = os.path.join(tmpdir, 'repo')
    restore_ok = False
    if bundle_present:
        try:
            # restore git repo bundle
            _run('git clone %s %s repo' % (cfg_params, repo_bundle_path), check=True, cwd=tmpdir)

            # restore original remote URL
            _run('git remote set-url origin %s' % repo_url, check=True, cwd=repo_dir)

            # pull latest stuff from remote
            _run('git pull', check=True, cwd=repo_dir)

            restore_ok = True
        except Exception:
            log.exception('cloning repo from bundle failed, skipping it')
            if os.path.exists(repo_dir):
                shutil.rmtree(repo_dir)

        # delete old bundle
        os.unlink(repo_bundle_path)

    # if restoring repo from bundle failed then do normal clone repo
    if not restore_ok:
        # clone repo
        #cmd = "git clone --single-branch --branch %s '%s' repo" % (repo_branch, repo_url)
        cmd = "git clone %s '%s' repo" % (cfg_params, repo_url)
        p = _run(cmd, check=False, cwd=tmpdir, capture_output=True, text=True)
        if p.returncode != 0:
            err = "command '%s' returned non-zero exit status %d\n" % (cmd, p.returncode)
            err += p.stdout.strip()[:140] + '\n'
            err += p.stderr.strip()[:140]
            err = err.strip()
            raise Exception(err)

    return repo_dir


def _collect_commits_from_git_log(text):
    commits = []
    commit = {'author': {}, 'added': [], 'modified': [], 'removed': []}
    record_lines = 0
    for line in text.splitlines():
        line = line.strip()
        # if empty line ie. record finished
        if not line:
            log.info('  %s', commit)
            commits.append(commit)
            commit = {'author': {}, 'added': [], 'modified': [], 'removed': []}
            record_lines = 0
            continue

        record_lines += 1
        if record_lines <= 5:
            field, val = line.split(':', 1)
            if field == 'commit':
                commit['id'] = val
            elif field == 'author':
                commit['author']['name'] = val
            elif field == 'email':
                commit['author']['email'] = val
            elif field == 'date':
                commit['timestamp'] = val
            elif field == 'subject':
                commit['message'] = val
            else:
                raise Exception('unrecognized field %s' % field)
        else:
            flag = line[0]
            fpath = line[1:].strip()
            if flag == 'A':
                commit['added'].append(fpath)
            elif flag == 'D':
                commit['removed'].append(fpath)
            else:
                commit['modified'].append(fpath)

    # add last non-empty entry
    if 'id' in commit:
        log.info('  %s', commit)
        commits.append(commit)

    return commits


def get_repo_commits_since(branch_id, prev_run, repo_url, repo_branch, git_cfg):
    log.info('checking commits in %s %s', repo_url, repo_branch)

    with tempfile.TemporaryDirectory(prefix='kraken-git-') as tmpdir:
        # prepare minio
        minio_bucket, minio_folder = minioops.get_or_create_minio_bucket_for_git(repo_url, branch_id=branch_id)
        mc = minioops.get_minio()
        minio_repo_bundle_path = '%s/repo.bundle' % minio_folder

        # retrieve repo from minio in bundle form or from remote git repository
        repo_dir = _retrieve_git_repo(tmpdir, repo_url, git_cfg, mc, minio_bucket, minio_repo_bundle_path)

        # get commits history
        cmd = "git log --no-merges --since='2 weeks ago' -n 20 --pretty=format:'commit:%H%nauthor:%an%nemail:%ae%ndate:%aI%nsubject:%s' --name-status"
        if prev_run and prev_run.repo_data_id and prev_run.repo_data.data:
            base_commit = None
            for rd in prev_run.repo_data.data:
                if rd['repo'] == repo_url:
                    base_commit = rd['commits'][0]['id']
                    break
            log.info('base commit: %s', base_commit)
            if base_commit:
                cmd += ' %s..' % base_commit
        else:
            base_commit = None
            log.info('no base commit %s', repo_url)
        p = _run(cmd, check=True, cwd=repo_dir, capture_output=True, text=True)
        text = p.stdout.strip()

        # bundle repo and cache it
        repo_bundle_path = os.path.join(tmpdir, 'repo.bundle')
        p = _run('git bundle create %s --all' % repo_bundle_path, check=False, cwd=repo_dir)
        if p.returncode != 0:
            log.warning('repo bundle failed, skipping it')
        else:
            log.info('store repo bundle %s -> %s / %s', repo_bundle_path, minio_bucket, minio_repo_bundle_path)
            try:
                mc.fput_object(minio_bucket, minio_repo_bundle_path, repo_bundle_path)
            except Exception:
                log.exception('problem with storing repo bundle, skipping it')

    # collect commits from git log
    commits = _collect_commits_from_git_log(text)

    return commits, base_commit


def get_schema_from_repo(repo_url, repo_branch, repo_access_token, schema_file, git_clone_params):  # pylint: disable=unused-argument
    with  tempfile.TemporaryDirectory(prefix='kraken-git-') as tmpdir:
        # clone repo
        if not git_clone_params:
            git_clone_params = ''

        if repo_access_token:
            url = furl.furl(repo_url)
            url.username = repo_access_token
            repo_url = url.tostr()

        cmd = "git clone --depth 1 --single-branch --branch %s %s '%s' repo" % (repo_branch, git_clone_params, repo_url)
        p = _run(cmd, check=False, cwd=tmpdir, capture_output=True, text=True, secret=repo_access_token)
        if p.returncode != 0:
            if repo_access_token:
                cmd = cmd.replace(repo_access_token, '******')
            err = "command '%s' returned non-zero exit status %d\n" % (cmd, p.returncode)
            err += p.stdout.strip()[:140] + '\n'
            err += p.stderr.strip()[:140]
            err = err.strip()
            raise Exception(err)

        repo_dir = os.path.join(tmpdir, 'repo')

        # get last commit SHA
        cmd = 'git rev-parse --verify HEAD'
        p = _run(cmd, check=True, cwd=repo_dir, capture_output=True, text=True)
        version = p.stdout.strip()

        # read schema code
        schema_path = os.path.join(repo_dir, schema_file)
        with open(schema_path, 'r') as f:
            schema_code = f.read()

    return schema_code, version


def clone_tool_repo(repo_url, repo_tag, tool_id):
    log.info('clone repo %s %s', repo_url, repo_tag)

    tmpdir = tempfile.TemporaryDirectory(prefix='kraken-git-')

    try:
        # prepare minio
        minio_bucket, minio_folder = minioops.get_or_create_minio_bucket_for_git(repo_url, tool_id=tool_id)
        mc = minioops.get_minio()
        minio_repo_bundle_path = '%s/repo.bundle' % minio_folder

        # retrieve repo from minio in bundle form or from remote git repository
        repo_dir = _retrieve_git_repo(tmpdir.name, repo_url, {}, mc, minio_bucket, minio_repo_bundle_path)

        # get last commit SHA
        cmd = 'git rev-parse --verify --short HEAD'
        p = _run(cmd, check=True, cwd=repo_dir, capture_output=True, text=True)
        version = p.stdout.strip()

        # bundle repo and cache it
        repo_bundle_path = os.path.join(tmpdir.name, 'repo.bundle')
        p = _run('git bundle create %s --all' % repo_bundle_path, check=False, cwd=repo_dir)
        if p.returncode != 0:
            log.warning('repo bundle failed, skipping it')
        else:
            log.info('store repo bundle %s -> %s / %s', repo_bundle_path, minio_bucket, minio_repo_bundle_path)
            try:
                mc.fput_object(minio_bucket, minio_repo_bundle_path, repo_bundle_path)
            except Exception:
                log.exception('problem with storing repo bundle, skipping it')
    except:
        tmpdir.cleanup()
        raise

    return tmpdir, os.path.join(tmpdir.name, 'repo'), version



"""added summary to flow

Revision ID: 5e38f55beb11
Revises: 899dadc28f9c
Create Date: 2022-03-26 14:08:23.398591

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '5e38f55beb11'
down_revision = '899dadc28f9c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('flows', sa.Column('summary', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('flows', 'summary')



"""added details to user

Revision ID: 411094f6b709
Revises: 0615faca43b1
Create Date: 2022-09-27 20:55:12.280567

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '411094f6b709'
down_revision = '0615faca43b1'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('users', sa.Column('details', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('users', 'details')



"""added users and users_session

Revision ID: d524cbcbf19c
Revises: 29b96da53f09
Create Date: 2021-02-03 07:34:18.936536

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd524cbcbf19c'
down_revision = '29b96da53f09'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('users',
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('updated', sa.DateTime(), nullable=False),
                    sa.Column('deleted', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=50), nullable=True),
                    sa.Column('password', sa.Unicode(length=150), nullable=True),
                    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('user_sessions',
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('updated', sa.DateTime(), nullable=False),
                    sa.Column('deleted', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('token', sa.Unicode(length=32), nullable=True),
                    sa.Column('user_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
                    sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('user_sessions')
    op.drop_table('users')



"""convert char(n) to text

Revision ID: ce05198f524a
Revises: b84bdc5d3c67
Create Date: 2021-06-26 08:33:25.798948

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ce05198f524a'
down_revision = 'b84bdc5d3c67'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('agents', 'address',
               existing_type=sa.VARCHAR(length=25),
               type_=sa.UnicodeText(),
               existing_nullable=False)
    op.alter_column('agents', 'ip_address',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('agents', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=False)
    op.alter_column('agents_groups', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('branches', 'branch_name',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('branches', 'name',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('files', 'path',
               existing_type=sa.VARCHAR(length=512),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('flows', 'branch_name',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('flows', 'label',
               existing_type=sa.VARCHAR(length=200),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('issues', 'message',
               existing_type=sa.VARCHAR(length=512),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('issues', 'path',
               existing_type=sa.VARCHAR(length=512),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('issues', 'symbol',
               existing_type=sa.VARCHAR(length=64),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('jobs', 'name',
               existing_type=sa.VARCHAR(length=200),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('jobs', 'notes',
               existing_type=sa.VARCHAR(length=2048),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('projects', 'description',
               existing_type=sa.VARCHAR(length=200),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('projects', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('runs', 'label',
               existing_type=sa.VARCHAR(length=200),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('secrets', 'name',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('settings', 'group',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('settings', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('settings', 'val_type',
               existing_type=sa.VARCHAR(length=8),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('stages', 'description',
               existing_type=sa.VARCHAR(length=1024),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('stages', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('systems', 'executor',
               existing_type=sa.VARCHAR(length=20),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('systems', 'name',
               existing_type=sa.VARCHAR(length=150),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('test_cases', 'name',
               existing_type=sa.VARCHAR(length=255),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('tools', 'description',
               existing_type=sa.VARCHAR(length=200),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('tools', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('user_sessions', 'token',
               existing_type=sa.VARCHAR(length=32),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('users', 'name',
               existing_type=sa.VARCHAR(length=50),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.alter_column('users', 'password',
               existing_type=sa.VARCHAR(length=150),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('users', 'password',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=150),
               existing_nullable=True)
    op.alter_column('users', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('user_sessions', 'token',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=32),
               existing_nullable=True)
    op.alter_column('tools', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('tools', 'description',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=200),
               existing_nullable=True)
    op.alter_column('test_cases', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)
    op.alter_column('systems', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=150),
               existing_nullable=True)
    op.alter_column('systems', 'executor',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=20),
               existing_nullable=True)
    op.alter_column('stages', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('stages', 'description',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=1024),
               existing_nullable=True)
    op.alter_column('settings', 'val_type',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=8),
               existing_nullable=True)
    op.alter_column('settings', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('settings', 'group',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('secrets', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)
    op.alter_column('runs', 'label',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=200),
               existing_nullable=True)
    op.alter_column('projects', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('projects', 'description',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=200),
               existing_nullable=True)
    op.alter_column('jobs', 'notes',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=2048),
               existing_nullable=True)
    op.alter_column('jobs', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=200),
               existing_nullable=True)
    op.alter_column('issues', 'symbol',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=64),
               existing_nullable=True)
    op.alter_column('issues', 'path',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=512),
               existing_nullable=True)
    op.alter_column('issues', 'message',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=512),
               existing_nullable=True)
    op.alter_column('flows', 'label',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=200),
               existing_nullable=True)
    op.alter_column('flows', 'branch_name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)
    op.alter_column('files', 'path',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=512),
               existing_nullable=True)
    op.alter_column('branches', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)
    op.alter_column('branches', 'branch_name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=255),
               existing_nullable=True)
    op.alter_column('agents_groups', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('agents', 'name',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=False)
    op.alter_column('agents', 'ip_address',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=50),
               existing_nullable=True)
    op.alter_column('agents', 'address',
               existing_type=sa.UnicodeText(),
               type_=sa.VARCHAR(length=25),
               existing_nullable=False)
    # ### end Alembic commands ###



# Copyright 2020-2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging
from urllib.parse import urljoin

from flask import abort
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.orm import joinedload

from . import consts
from . import utils
from .models import db, Branch, Flow, Run, Stage, Job, Step
from .models import Issue, Artifact, Agent
from .schema import SchemaError
from . import exec_utils
from . import access
from . import chops

log = logging.getLogger(__name__)


def create_flow(branch_id, kind, body, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch %s not found" % branch_id)
    access.check(token_info, branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can create a flow')

    log.set_ctx(branch=branch.id)

    try:
        flow = exec_utils.create_a_flow(branch, kind, body)
    except SchemaError as e:
        abort(400, str(e))

    data = flow.get_json()

    return data, 201


def get_flows(branch_id, kind, start=0, limit=10, middle=None, token_info=None):
    branch = Branch.query.filter_by(id=branch_id).one_or_none()
    if branch is None:
        abort(404, "Branch %s not found" % branch_id)
    access.check(token_info, branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get flows')

    t0 = time.time()

    flows = []
    if kind == 'dev':
        kind = 1
    else:
        kind = 0
    q = Flow.query.filter_by(branch_id=branch_id, kind=kind)
    q = q.options(joinedload('branch'),
                  # joinedload('branch.project'),
                  joinedload('branch.stages'),
                  joinedload('artifacts_files'),
                  joinedload('runs'))
                  # joinedload('runs.artifacts_files'))
    if middle is None:
        total = q.count()
        q = q.order_by(desc(Flow.created))
        q = q.offset(start).limit(limit)
        for flow in q.all():
            flows.append(flow.get_json(with_project=False, with_branch=False, with_schema=False))

    else:
        q1 = q.filter(Flow.id >= middle)
        q1 = q1.order_by(asc(Flow.created))
        q1 = q1.offset(0).limit(limit)
        for flow in reversed(q1.all()):
            flows.append(flow.get_json(with_project=False, with_branch=False, with_schema=False))

        q2 = q.filter(Flow.id < middle)
        q2 = q2.order_by(desc(Flow.created))
        q2 = q2.offset(0).limit(limit)
        for flow in q2.all():
            flows.append(flow.get_json(with_project=False, with_branch=False, with_schema=False))

        total = 0

    log.info('get_flows time %s', time.time() - t0)

    return {'items': flows, 'total': total}, 200


def get_flow(flow_id, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow not found")
    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get a flow')

    return flow.get_json()


def get_flow_runs(flow_id, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow not found")
    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get flow runs')

    runs = []
    for run in flow.runs:
        runs.append(run.get_json())
    return runs, 200


def get_flow_artifacts(flow_id, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow not found")
    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get flow artifacts')

    base_url = '/bk/artifacts/public/f/%d/' % flow_id

    artifacts = []
    for art in Artifact.query.filter_by(flow=flow, section=consts.ARTIFACTS_SECTION_PUBLIC):
        art = art.get_json()
        art['url'] = urljoin(base_url, art['path'].strip('/'))
        artifacts.append(art)
    return {'items': artifacts, 'total': len(artifacts)}, 200


def create_run(flow_id, body, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow not found")
    access.check(token_info, flow.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can create a run')

    stage = Stage.query.filter_by(id=body['stage_id']).one_or_none()
    if stage is None:
        abort(404, "Stage not found")

    log.set_ctx(branch=flow.branch_id, flow_kind=flow.kind, flow=flow_id)

    run = exec_utils.start_run(stage, flow, reason=dict(reason='manual'), args=body.get('args', {}))

    # Serialize and return the newly created run in the response
    data = run.get_json()
    return data, 201


def run_run_jobs(run_id, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can run jobs')

    log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

    if run.state == consts.RUN_STATE_MANUAL:
        replay = False
    else:
        replay = True

    try:
        exec_utils.trigger_jobs(run, replay=replay)
    except SchemaError as e:
        abort(400, str(e))

    data = run.get_json()

    return data, 200


def job_rerun(job_id, token_info=None):
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")
    access.check(token_info, job.run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can rerun a job')

    log.set_ctx(branch=job.run.flow.branch_id, flow_kind=job.run.flow.kind, flow=job.run.flow_id, run=job.run.id)

    # TODO rerun
    job2 = Job(run=job.run, name=job.name, agents_group=job.agents_group, system=job.system, timeout=job.timeout)

    for s in job.steps:
        Step(job=job2, index=s.index, tool=s.tool, fields=s.fields)

    job.covered = True
    job2.run.state = consts.RUN_STATE_IN_PROGRESS  # need to be set in case of replay
    db.session.commit()

    data = job2.get_json()
    return data, 200


def create_job(job, token_info=None):
    run_id = job.get("run")
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can create a job')

    # TODO
    # schema = JobSchema()
    # new_job = schema.load(job, session=db.session).data
    # db.session.commit()

    # # Serialize and return the newly created person in the response
    # data = schema.dump(new_job).data
    data = {}

    return data, 201


def get_runs(stage_id, token_info=None):
    stage = Stage.query.filter_by(id=stage_id).one_or_none()
    if stage is None:
        abort(404, "Stage not found")
    access.check(token_info, stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get runs')

    q = Run.query.filter_by(stage_id=stage_id)
    runs = []
    for run in q.all():
        runs.append(run.get_json())
    return runs, 200


def get_run_jobs(run_id, start=0, limit=10, include_covered=False, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get run jobs')

    q = Job.query
    q = q.filter_by(run_id=run_id)
    if not include_covered:
        q = q.filter_by(covered=False)
    total = q.count()
    q = q.order_by(asc('id'))
    q = q.offset(start).limit(limit)
    jobs = []
    for j in q.all():
        js = j.get_json(mask_secrets=True)
        jobs.append(js)
    return {'items': jobs, 'total': total}, 200


def get_run_issues(run_id, start=0, limit=10, issue_types=None, location=None, message=None,
                   symbol=None, min_age=None, max_age=None, job=None, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get run issues')

    q = Issue.query
    q = q.options(joinedload('job'),
                  joinedload('job.agents_group'),
                  joinedload('job.agent_used'))
    q = q.join('job')
    q = q.filter(Job.run_id == run_id, Job.covered.is_(False))
    if issue_types:
        q = q.filter(Issue.issue_type.in_(issue_types))
    if location is not None:
        q = q.filter(Issue.path.ilike('%' + location + '%'))
    if message is not None:
        q = q.filter(Issue.message.ilike('%' + message + '%'))
    if symbol is not None:
        q = q.filter(Issue.symbol.ilike('%' + symbol + '%'))
    if min_age is not None:
        q = q.filter(Issue.age >= min_age)
    if max_age is not None:
        q = q.filter(Issue.age <= max_age)
    if job is not None:
        if job.isdigit():
            job_id = int(job)
            q = q.filter(Job.id == job_id)
        else:
            q = q.filter(Job.name.ilike('%' + job + '%'))

    total = q.count()
    q = q.order_by(asc('path'), asc('line'))
    q = q.offset(start).limit(limit)
    issues = []
    for i in q.all():
        issues.append(i.get_json())
    return {'items': issues, 'total': total}, 200


def get_run_artifacts(run_id, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get run artifacts')

    base_url = '/bk/artifacts/public/r/%d/' % run.id

    artifacts = []
    for art in Artifact.query.filter_by(run=run, section=consts.ARTIFACTS_SECTION_PUBLIC):
        art = art.get_json()
        art['url'] = urljoin(base_url, art['path'].strip('/'))
        artifacts.append(art)
    return {'items': artifacts, 'total': len(artifacts)}, 200


def get_run(run_id, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get a run')

    return run.get_json(), 200


def get_job_logs(job_id, start=0, limit=200, order=None, internals=False, filters=None, token_info=None):  # pylint: disable=unused-argument
    if order not in [None, 'asc', 'desc']:
        abort(400, "incorrect order value: %s" % str(order))

    if start < 0:
        abort(400, "incorrect start value: %s" % str(start))

    if limit < 0:
        abort(400, "incorrect limit value: %s" % str(limit))

    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")
    access.check(token_info, job.run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get job logs')

    job_json = job.get_json()

    ch = chops.get_clickhouse()

    internal_clause = ''
    if not internals:
        internal_clause = "and tool != '' and service = 'agent'"

    query = "select count(*) from logs where job = %%(job_id)d %s" % internal_clause
    params = dict(job_id=job_id)
    resp = ch.execute(query, params)
    total = resp[0][0]

    if order is None:
        order = 'asc'

    query = "select time,message,service,host,level,job,tool,step from logs where job = %%(job_id)d %s order by time %s, seq %s limit %%(start)d, %%(limit)d"
    query %= (internal_clause, order, order)
    params = dict(job_id=job_id, start=start, limit=limit)

    rows = ch.execute(query, params)

    logs = []
    for r in rows:
        entry = dict(time=r[0],
                     message=r[1],
                     service=r[2],
                     host=r[3],
                     level=r[4].lower()[:4],
                     job=r[5],
                     tool=r[6],
                     step=r[7])
        logs.append(entry)

    return {'items': logs, 'total': total, 'job': job_json}, 200


def get_job(job_id, token_info=None):
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")
    access.check(token_info, job.run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get a job')

    return job.get_json()


def get_step_logs(job_id, step_idx, start=0, limit=200, order=None, filters=None, token_info=None):  # pylint: disable=unused-argument
    if order not in [None, 'asc', 'desc']:
        abort(400, "incorrect order value: %s" % str(order))

    if start < 0:
        abort(400, "incorrect start value: %s" % str(start))

    if limit < 0:
        abort(400, "incorrect limit value: %s" % str(limit))

    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")
    access.check(token_info, job.run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power and project viewer user roles can get job logs')

    job_json = job.get_json()

    ch = chops.get_clickhouse()

    query = "SELECT count(*) FROM logs WHERE job = %(job_id)d AND step = %(step_idx)d"
    params = dict(job_id=job_id, step_idx=step_idx)
    resp = ch.execute(query, params)
    total = resp[0][0]

    if order is None:
        order = 'asc'

    query = "SELECT time,message,service,host,level,job,tool,step "
    query += " FROM logs WHERE job = %%(job_id)d AND step = %%(step_idx)d "
    query += " ORDER BY time %s, seq %s LIMIT %%(start)d, %%(limit)d"
    query %= (order, order)
    params = dict(job_id=job_id, step_idx=step_idx, start=start, limit=limit)

    rows = ch.execute(query, params)

    logs = []
    for r in rows:
        entry = dict(time=r[0],
                     message=r[1],
                     service=r[2],
                     host=r[3],
                     level=r[4].lower()[:4],
                     job=r[5],
                     tool=r[6],
                     step=r[7])
        logs.append(entry)

    return {'items': logs, 'total': total, 'job': job_json}, 200


def get_logs(branch_id=None, flow_kind=None, flow_id=None, run_id=None, job_id=None, step_idx=None,
             agent_id=None, services=None, level=None,
             start=0, limit=200, order=None, token_info=None):  # pylint: disable=unused-argument
    if order not in [None, 'asc', 'desc']:
        abort(400, "incorrect order value: %s" % str(order))

    if start < 0:
        abort(400, "incorrect start value: %s" % str(start))

    if limit < 0:
        abort(400, "incorrect limit value: %s" % str(limit))

    if branch_id is None and flow_id is None and run_id is None and job_id is None and agent_id is None:
        access.check(token_info, '', 'admin',
                     'only superadmin can get services logs')

    if branch_id is None and flow_kind is not None:
        abort(400, "if flow_kind is provided then branch_id must be provided as well")

    if job_id is None and step_idx is not None:
        abort(400, "if step_idx is provided then job_id must be provided as well")

    if job_id is not None:
        job = Job.query.filter_by(id=job_id).one_or_none()
        if job is None:
            abort(404, "Job not found")
        access.check(token_info, job.run.stage.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get job logs')

    elif run_id is not None:
        run = Run.query.filter_by(id=run_id).one_or_none()
        if run is None:
            abort(404, "Run not found")
        access.check(token_info, run.stage.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get run logs')

    elif flow_id is not None:
        flow = Flow.query.filter_by(id=flow_id).one_or_none()
        if flow is None:
            abort(404, "Flow not found")
        access.check(token_info, flow.branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get flow logs')

    elif branch_id is not None:
        branch = Branch.query.filter_by(id=branch_id).one_or_none()
        if branch is None:
            abort(404, "Branch not found")
        access.check(token_info, branch.project_id, 'view',
                     'only superadmin, project admin, project power and project viewer user roles can get branch logs')

    if agent_id is not None:
        agent = Agent.query.filter_by(id=agent_id).one_or_none()
        if agent is None:
            abort(404, "Agent not found")
        access.check(token_info, '', 'admin',
                     'only superadmin can get agent logs')

    ch = chops.get_clickhouse()

    if order is None:
        order = 'asc'

    params = dict(start=start, limit=limit)

    where_clauses = []
    if branch_id is not None:
        where_clauses.append('branch = %(branch_id)d')
        params['branch_id'] = branch_id
        if flow_kind is not None:
            where_clauses.append('flow_kind = %(flow_kind)d')
            params['flow_kind'] = flow_kind
    if flow_id is not None:
        where_clauses.append('flow = %(flow_id)d')
        params['flow_id'] = flow_id
    if run_id is not None:
        where_clauses.append('run = %(run_id)d')
        params['run_id'] = run_id
    if job_id is not None:
        where_clauses.append('job = %(job_id)d')
        params['job_id'] = job_id
        if step_idx is not None:
            where_clauses.append('step_idx = %(step_idx)d')
            params['step_idx'] = step_idx
    if agent_id is not None:
        where_clauses.append('agent = %(agent_id)d')
        params['agent_id'] = agent_id

    where_services = []
    if services:
        for idx, s in enumerate(services):
            param = 'service%d' % idx
            if '/' in s:
                s, t = s.split('/')
                tparam = 'tool%d' % idx
                where_services.append("(service = %%(%s)s and tool = %%(%s)s)" % (param, tparam))
                params[param] = s
                params[tparam] = t
            else:
                where_services.append("service = %%(%s)s" % param)
                params[param] = s
        where = " or ".join(where_services)
        where = "(" + where + ") "
        where_clauses.append(where)

    if level:
        level = level.upper()
        if level == 'ERROR':
            lq = "level = 'ERROR'"
        elif level == 'WARNING':
            lq = "level in ('WARNING', 'ERROR')"
        else:
            lq = "level in ('INFO', 'WARNING', 'ERROR')"
        where_clauses.append(lq)

    where_clause = 'where ' + ' and '.join(where_clauses)

    # get total first
    total_query = "SELECT count(*) FROM logs %s" % where_clause
    resp = ch.execute(total_query, params)
    total = resp[0][0]

    # get logs now
    query = "SELECT time,message,service,host,path,lineno,level,branch,flow_kind,flow,run,job,tool,step,agent "
    query += " FROM logs %s ORDER BY time %s, seq %s LIMIT %%(start)d, %%(limit)d"
    query %= (where_clause, order, order)
    log.info('CH query %s', query)
    log.info('CH params %s', params)
    rows = ch.execute(query, params)

    logs = []
    for r in rows:
        entry = dict(time=r[0],
                     message=r[1],
                     service=r[2],
                     host=r[3],
                     path=r[4],
                     lineno=r[5],
                     level=r[6].lower()[:4],
                     branch=r[7],
                     flow_kind=r[8],
                     flow=r[9],
                     run=r[10],
                     job=r[11],
                     tool=r[12],
                     step=r[13],
                     agent=r[14])
        logs.append(entry)

    return {'items': logs, 'total': total}, 200


def cancel_run(run_id, token_info=None):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run not found")
    access.check(token_info, run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can cancel a run')

    # if run is completed then do nothing
    if run.state == consts.RUN_STATE_COMPLETED:
        return {}

    log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

    # cancel any pending job
    all_completed = True
    for job in run.jobs:
        if job.state != consts.JOB_STATE_COMPLETED:
            log.set_ctx(job=job.id)
            exec_utils.cancel_job(job, 'canceled by user', consts.JOB_CMPLT_USER_CANCEL)
            all_completed = False
        log.set_ctx(job=None)

    # if all jobs already completed or there is no jobs then complete run
    if all_completed:
        now = utils.utcnow()
        exec_utils.complete_run(run, now)

    run.note = 'run %d canceled by user' % run.id
    db.session.commit()

    return {}


def delete_job(job_id, token_info=None):
    job = Job.query.filter_by(id=job_id).one_or_none()
    if job is None:
        abort(404, "Job not found")
    access.check(token_info, job.run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power roles can cancel a job')

    log.set_ctx(branch=job.run.flow.branch_id, flow_kind=job.run.flow.kind, flow=job.run.flow_id, run=job.run.id, job=job.id)

    exec_utils.cancel_job(job, 'canceled by user', consts.JOB_CMPLT_USER_CANCEL)

    return {}


def get_agent_jobs(agent_id, start=0, limit=10, token_info=None):
    access.check(token_info, '', 'admin',
                 'only superadmin role can get agent jobs')
    q = Job.query
    q = q.filter_by(agent_used_id=agent_id)
    total = q.count()
    q = q.order_by(desc('id'))
    q = q.offset(start).limit(limit)
    jobs = []
    for j in q.all():
        jobs.append(j.get_json())
    return {'items': jobs, 'total': total}, 200



"""removed enabled from projects

Revision ID: 1f5039135dd8
Revises: ab9326d5fc02
Create Date: 2021-06-20 11:17:17.259316

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '1f5039135dd8'
down_revision = 'ab9326d5fc02'
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        try:
            op.drop_column('projects', 'enabled')
        except Exception:
            pass

    with op.get_context().autocommit_block():
        try:
            op.drop_constraint("systems_executor_key", "systems", type_="unique")
        except Exception:
            pass

    with op.get_context().autocommit_block():
        try:
            op.drop_constraint("systems_name_key", "systems", type_="unique")
        except Exception:
            pass


    with op.get_context().autocommit_block():
        try:
            op.drop_column('flows', 'trigger_data')
        except Exception:
            pass

    with op.get_context().autocommit_block():
        try:
            op.drop_column('runs', 'repo_data')
        except Exception:
            pass

    with op.get_context().autocommit_block():
        try:
            op.alter_column('jobs', 'system_id',
                            nullable=False,
                            existing_nullable=True)
        except Exception:
            pass


def downgrade():
    op.add_column('projects', sa.Column('enabled', sa.Boolean))



"""added missing indexes

Revision ID: 899dadc28f9c
Revises: a0741baf2dbc
Create Date: 2022-01-20 07:45:48.954451

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '899dadc28f9c'
down_revision = 'a0741baf2dbc'
branch_labels = None
depends_on = None

INDEXES = [
    ('ix_artifacts_flow_id', 'artifacts', 'flow_id'),
    ('ix_artifacts_run_id', 'artifacts', 'run_id'),
    ('ix_flows_branch_id_kind', 'flows', 'branch_id, kind'),
    ('ix_flows_created', 'flows', 'created'),
    ('ix_runs_flow_id', 'runs', 'flow_id'),
    ('uq_system_name_executor', 'systems', 'name, executor'),
    ('ix_test_case_results_comment_id', 'test_case_results', 'comment_id'),
    ('ix_test_case_results_job_id', 'test_case_results', 'job_id')
]


def upgrade():
    conn = op.get_bind()
    for name, table, columns in INDEXES:
        if name.startswith('uq'):
            uq = ' UNIQUE '
            cmd = 'ALTER TABLE %s ADD CONSTRAINT %s UNIQUE (%s)' % (table, name, columns)
            with op.get_context().autocommit_block():
                try:
                    conn.execute(cmd)
                except Exception:
                    pass
        else:
            uq = ''
        cmd = "CREATE %s INDEX IF NOT EXISTS %s ON public.%s USING btree (%s);" % (uq, name, table, columns)
        print(cmd)
        conn.execute(cmd)
    print('migration completed')


def downgrade():
    conn = op.get_bind()
    for name, table, _ in INDEXES:
        if name.startswith('uq'):
            cmd = 'ALTER TABLE %s DROP CONSTRAINT %s' % (table, name)
            conn.execute(cmd)
        print('dropping index %s' % name)
        cmd = "DROP INDEX IF EXISTS %s;" % name
        conn.execute(cmd)
        #op.drop_index(name, table_name=table)
    print('migration completed')



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.sql.expression import desc

from . import consts
from .models import db, BranchSequence, Flow, Run, Job, Step, AgentsGroup, Tool, Agent, AgentAssignment, System
from .schema import check_and_correct_stage_schema, prepare_secrets, substitute_vars, substitute_val
from . import dbutils
from . import kkrq
from . import utils

log = logging.getLogger(__name__)


def complete_run(run, now):
    from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
    run.state = consts.RUN_STATE_COMPLETED
    run.finished = now
    db.session.commit()
    log.info('completed run %s, now: %s', run, run.finished)

    # trigger run analysis
    kkrq.enq(bg_jobs.analyze_run, run.id)
    log.info('run %s completed', run)


def cancel_job(job, note, cmplt_status):
    # job is not un-assigned from agent, this will happen in backend.py
    # when agent will call get_job or any other function
    # or in watchdog.py in _check_agents_keep_alive if agent is not alive
    from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
    if job.state == consts.JOB_STATE_COMPLETED:
        return

    log.set_ctx(job=job.id)

    job.completed = utils.utcnow()
    job.state = consts.JOB_STATE_COMPLETED
    job.completion_status = cmplt_status
    job.notes = note
    job.finished = utils.utcnow()
    db.session.commit()
    kkrq.enq(bg_jobs.job_completed, job.id)
    log.info('job %s canceled because: %s', job, note, job=job.id)


def _increment_sequences(branch, stage, kind):
    if stage is None:
        if kind == 0:  # CI
            seq1_kind = consts.BRANCH_SEQ_FLOW
            seq1_name = 'KK_FLOW_SEQ'
            seq2_kind = consts.BRANCH_SEQ_CI_FLOW
        else:
            seq1_kind = consts.BRANCH_SEQ_FLOW
            seq1_name = 'KK_FLOW_SEQ'
            seq2_kind = consts.BRANCH_SEQ_CI_FLOW
        seq2_name = 'KK_CI_DEV_FLOW_SEQ'
    else:
        if kind == 0:  # CI
            seq1_kind = consts.BRANCH_SEQ_RUN
            seq1_name = 'KK_RUN_SEQ'
            seq2_kind = consts.BRANCH_SEQ_DEV_RUN
        else:
            seq1_kind = consts.BRANCH_SEQ_RUN
            seq1_name = 'KK_RUN_SEQ'
            seq2_kind = consts.BRANCH_SEQ_DEV_RUN
        seq2_name = 'KK_CI_DEV_RUN_SEQ'

    seq1 = BranchSequence.query.filter_by(branch=branch, stage=stage, kind=seq1_kind).one()
    seq1.value = BranchSequence.value + 1
    seq2 = BranchSequence.query.filter_by(branch=branch, stage=stage, kind=seq2_kind).one()
    seq2.value = BranchSequence.value + 1
    db.session.commit()

    vals = {}
    vals[seq1_name] = str(seq1.value)
    vals[seq2_name] = str(seq2.value)
    return vals


def complete_starting_run(run_id):
    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        raise Exception('run %d cannot be found' % run_id)

    log.set_ctx(branch=run.flow.branch_id, flow_kind=run.flow.kind, flow=run.flow_id, run=run.id)

    log.info('complete starting run: %s', run)

    # if this is already executed run then replay it, if this is new, fresh run then do everything from scratch
    if run.state == consts.RUN_STATE_REPLAY:
        # move run back into IN PROGRESS state
        run.state = consts.RUN_STATE_IN_PROGRESS
        replay = True
    else:
        replay = False

        # prepare args
        run_args = run.stage.get_default_args()
        if run.flow.args:
            run_args.update(run.flow.args)
        if run.args is not None:
            run_args.update(run.args)

        # increment sequences
        seq_vals = _increment_sequences(run.flow.branch, run.stage, run.flow.kind)
        run_args.update(seq_vals)

        # prepare flow label if needed
        lbl_vals = {}
        for lbl_field in ['flow_label', 'run_label']:
            label_pattern = run.stage.schema.get(lbl_field, None)
            if label_pattern:
                lbl_vals[lbl_field] = label_pattern
        if lbl_vals:
            lbl_vals, _ = substitute_vars(lbl_vals, run_args)
            if run.flow.label is None:
                run.flow.label = lbl_vals.get('flow_label', None)

        # if currently there is not repo data copy it from previous run if it is there
        repo_data = run.repo_data
        if not repo_data:
            prev_run = dbutils.get_prev_run(run.stage.id, run.flow.kind)
            if prev_run and prev_run.repo_data_id and prev_run.repo_data.data:
                repo_data = prev_run.repo_data
                log.info('new run, taken repo_data from prev run %s', prev_run)

        # initialize run instance
        run.args = run_args
        run.label = lbl_vals.get('run_label', None)
        run.repo_data = repo_data

        # move run from CREATING to proper state
        if run.stage.schema['triggers'].get('manual', False):
            run.state = consts.RUN_STATE_MANUAL
        else:
            run.state = consts.RUN_STATE_IN_PROGRESS

    db.session.commit()

    # trigger jobs if not manual
    if run.state == consts.RUN_STATE_MANUAL:
        log.info('created manual run %s for stage %s of branch %s - no jobs started yet', run, run.stage, run.stage.branch)
    else:
        # update pointer to last, incomplete, CI flow in the branch
        if run.flow.kind == consts.FLOW_KIND_CI:
            last_flow = run.flow.branch.ci_last_incomplete_flow
            if last_flow is None or last_flow.created < run.flow.created:
                run.flow.branch.ci_last_incomplete_flow = run.flow

        # trigger jobs
        log.info('starting run %s for stage %s of branch %s', run, run.stage, run.stage.branch)
        # TODO: move triggering jobs to background tasks
        trigger_jobs(run, replay=replay)

    return run


def start_run(stage, flow, reason, args=None, repo_data=None):
    # if there was already run for this stage then replay it, otherwise create new one
    run = Run.query.filter_by(stage=stage, flow=flow).one_or_none()
    if run is None:
        # create run instance
        run = Run(stage=stage, flow=flow, args=args, reason=reason, repo_data=repo_data, state=consts.RUN_STATE_CREATING)
    else:
        # move run to REPLAY state
        run.state = consts.RUN_STATE_REPLAY
    db.session.commit()

    log.set_ctx(run=run.id)

    if stage.schema_from_repo_enabled:
        from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel
        kkrq.enq_neck(bg_jobs.refresh_schema_repo, stage.id, run.id, ignore_args=[1])
    else:
        complete_starting_run(run.id)

    return run


def create_a_flow(branch, kind, body, trigger_data=None):
    if kind == 'dev':
        kind = 1
    else:
        kind = 0

    args = body.get('args', {})
    flow_args = args.get('Common', {})
    branch_name = flow_args.get('BRANCH', branch.branch_name)
    if 'BRANCH' in flow_args:
        del flow_args['BRANCH']

    # increment sequences
    seq_vals = _increment_sequences(branch, None, kind)
    flow_args.update(seq_vals)

    flow_args['KK_FLOW_TYPE'] = 'CI' if kind == 0 else 'DEV'
    flow_args['KK_BRANCH'] = branch_name if branch_name else 'master'

    # create flow instance
    flow = Flow(branch=branch, kind=kind, branch_name=branch_name, args=flow_args, trigger_data=trigger_data)
    db.session.commit()

    log.set_ctx(flow_kind=flow.kind, flow=flow.id)

    log.info('created %s flow %s in branch %s',
             'ci' if kind == 0 else 'dev',
             flow, branch)

    reason = dict(reason='manual')

    for stage in branch.stages:
        if stage.deleted:
            continue
        if stage.schema['parent'] != 'root' or stage.schema['triggers'].get('parent', True) is False:
            continue

        if not stage.enabled:
            log.info('stage %s not started - disabled', stage)
            continue

        start_run(stage, flow, reason=reason, args=args.get(stage.name, {}))

    return flow


def _setup_schema_context(run):
    ctx = {
        'is_ci': run.flow.kind == 0,
        'is_dev': run.flow.kind == 1,
        'run_label': run.label,
        'flow_label': run.flow.label,
    }
    return ctx


def _reeval_schema(run):
    context = _setup_schema_context(run)

    schema_code, schema = check_and_correct_stage_schema(run.stage.branch, run.stage.name, run.stage.schema_code, context)

    run.stage.schema = schema
    run.stage.schema_code = schema_code
    flag_modified(run.stage, 'schema')
    db.session.commit()


def _find_covered_jobs(run):
    covered_jobs = {}
    q = Job.query.filter_by(run=run).filter_by(covered=False)
    for j in q.all():
        key = '%s-%s' % (j.name, j.agents_group_id)
        if key not in covered_jobs:
            covered_jobs[key] = [j]
        else:
            covered_jobs[key].append(j)

    return covered_jobs


def _establish_timeout_for_job(j, run, system, agents_group):
    if agents_group is not None:
        job_key = "%s-%d-%d" % (j['name'], system.id, agents_group.id)
        if run.stage.timeouts and job_key in run.stage.timeouts:
            # take estimated timeout if present
            timeout = run.stage.timeouts[job_key]
        else:
            # take initial timeout from schema, or default one
            timeout = int(j.get('timeout', consts.DEFAULT_JOB_TIMEOUT))
            timeout = max(timeout, 60)
    else:
        timeout = consts.DEFAULT_JOB_TIMEOUT

    return timeout


def trigger_jobs(run, replay=False):
    log.info('triggering jobs for run %s', run)

    # reevaluate schema code
    _reeval_schema(run)

    schema = run.stage.schema

    # find any prev jobs that will be covered by jobs triggered here in this function
    if replay:
        covered_jobs = _find_covered_jobs(run)

    # prepare secrets to pass them to substitute in steps
    secrets = prepare_secrets(run)

    # prepare missing group
    missing_agents_group = AgentsGroup.query.filter_by(name='missing').one_or_none()
    if missing_agents_group is None:
        missing_agents_group = AgentsGroup(name='missing')
        db.session.commit()

    # count how many agents are in each group, if there is 0 for given job then return an error
    agents_count = {}

    agents_needed = set()
    created_systems = {}

    # trigger new jobs based on jobs defined in stage schema
    all_started_erred = True
    now = utils.utcnow()
    for j in schema['jobs']:
        # check tools in steps
        tools = []
        tool_not_found = None
        for idx, s in enumerate(j['steps']):
            if '@' in s['tool']:
                name, ver = s['tool'].split('@')
                tool = Tool.query.filter_by(name=name, version=ver).one_or_none()
            else:
                tool = Tool.query.filter_by(name=s['tool']).order_by(desc(Tool.created)).first()
            if tool is None:
                tool_not_found = s['tool']
                break
            tools.append(tool)
        if tool_not_found is not None:
            log.warning('cannot find tool %s', tool_not_found)

        envs = j['environments']
        for env in envs:
            # get agents group
            ag_name, _ = substitute_val(env['agents_group'], run.args)
            q = AgentsGroup.query
            q = q.filter_by(project=run.stage.branch.project, name=ag_name)
            agents_group = q.one_or_none()

            if agents_group is None:
                agents_group = AgentsGroup.query.filter_by(name=ag_name).one_or_none()
                if agents_group is None:
                    log.warning("cannot find agents group '%s'", ag_name)

            # get count of agents in the group
            if agents_group is not None and agents_group.name not in agents_count:
                q = AgentAssignment.query.filter_by(agents_group=agents_group)
                q = q.join('agent')
                q = q.filter(Agent.disabled.is_(False))
                q = q.filter(Agent.authorized.is_(True))
                cnt = q.count()
                #log.info("agents group '%s' count is %d", agents_group.name, cnt)
                agents_count[agents_group.name] = cnt

            if not isinstance(env['system'], list):
                systems = [substitute_val(env['system'], run.args)[0]]
            else:
                systems = [substitute_val(s, run.args)[0] for s in env['system']]

            for system_name in systems:
                # prepare system and executor
                if 'executor' in env:
                    executor = env['executor'].lower()
                else:
                    executor = 'local'
                system = System.query.filter_by(name=system_name, executor=executor).one_or_none()
                sys_key = (system_name, executor)
                if system is None:
                    system = created_systems.get(sys_key, None)
                if system is None:
                    system = System(name=system_name, executor=executor)
                    db.session.flush()
                    # this is to avoid doing flush for the same system
                    created_systems[sys_key] = system

                # get timeout
                timeout = _establish_timeout_for_job(j, run, system, agents_group)

                # create job
                job = Job(run=run, name=j['name'], agents_group=agents_group, system=system, timeout=timeout)

                erred_job = False
                if tool_not_found:
                    job.state = consts.JOB_STATE_COMPLETED
                    job.completion_status = consts.JOB_CMPLT_MISSING_TOOL_IN_DB
                    job.notes = "cannot find tool '%s' in database" % tool_not_found
                    erred_job = True
                if agents_group is None:
                    job.agents_group = missing_agents_group
                    if job.state != consts.JOB_STATE_COMPLETED:
                        job.state = consts.JOB_STATE_COMPLETED
                        job.completion_status = consts.JOB_CMPLT_MISSING_AGENTS_GROUP
                        job.notes = "cannot find agents group '%s' in database" % ag_name
                    erred_job = True
                elif agents_group.deployment and agents_group.deployment['method'] > 0:
                    agents_needed.add(agents_group.id)
                elif agents_count[agents_group.name] == 0:
                    if job.state != consts.JOB_STATE_COMPLETED:
                        job.state = consts.JOB_STATE_COMPLETED
                        job.completion_status = consts.JOB_CMPLT_NO_AGENTS
                        job.notes = "there are no agents in group '%s' - add some agents" % agents_group.name
                    erred_job = True

                if not erred_job:
                    all_started_erred = False
                    # substitute vars in steps
                    for idx, s in enumerate(j['steps']):
                        args = secrets.copy()
                        if run.args:
                            args.update(run.args)
                        fields, fields_masked = substitute_vars(s, args)
                        del fields['tool']
                        Step(job=job, index=idx, tool=tools[idx], fields=fields, fields_masked=fields_masked)

                # if this is rerun/replay then mark prev jobs as covered
                if replay:
                    key = '%s-%s' % (j['name'], agents_group.id)
                    if key in covered_jobs:
                        for cj in covered_jobs[key]:
                            cj.covered = True
                            # TODO: we should cancel these jobs if they are still running

                db.session.commit()
                log.info('created job %s', job.get_json(mask_secrets=True))

    run.started = now
    run.state = consts.RUN_STATE_IN_PROGRESS  # need to be set in case of replay
    db.session.commit()

    from .bg import jobs as bg_jobs  # pylint: disable=import-outside-toplevel

    # spawn new agents if needed
    if agents_needed:
        for ag_id in agents_needed:
            kkrq.enq_neck(bg_jobs.spawn_new_agents, ag_id)
        log.info('enqueued spawning new agents for run %s', run)

    # notify
    kkrq.enq(bg_jobs.notify_about_started_run, run.id)
    log.info('enqueued notification about start of run %s', run)

    if len(schema['jobs']) == 0 or all_started_erred:
        complete_run(run, now)



"""added last_seen to Executor

Revision ID: ff4a425a6070
Revises: 735e81eedae4
Create Date: 2020-01-13 07:24:56.514075

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'ff4a425a6070'
down_revision = '735e81eedae4'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('executors', sa.Column('last_seen', sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column('executors', 'last_seen')
